import tarfile
import tempfile
from pathlib import Path
from unittest.mock import patch

from lxml import etree
from parameterized import parameterized

from sdlon.sd_log_analyzer import extract_log_file_lines
from sdlon.sd_log_analyzer import get_all_sd_person_changes
from sdlon.sd_log_analyzer import get_sd_person_changed
from sdlon.sd_log_analyzer import get_sd_person_changed_at_date_responses
from sdlon.sd_log_analyzer import get_sd_xml_responses
from sdlon.sd_log_analyzer import get_tar_gz_archive_files
from sdlon.sd_log_analyzer import IdType
from sdlon.sd_log_analyzer import output_to_file
from sdlon.sd_log_analyzer import SdPersonChange


remove_blank_parser = etree.XMLParser(remove_blank_text=True)

XML_FIXTURE = etree.XML(
    """
    <GetPersonChangedAtDate20111201>
        <RequestStructure>
            <InstitutionIdentifier>XY</InstitutionIdentifier>
            <ActivationDate>2021-09-14</ActivationDate>
            <ActivationTime>00:00:00</ActivationTime>
            <DeactivationDate>2021-09-15</DeactivationDate>
            <DeactivationTime>00:00:59</DeactivationTime>
            <ContactInformationIndicator>false</ContactInformationIndicator>
            <PostalAddressIndicator>false</PostalAddressIndicator>
        </RequestStructure>
        <Person>
            <PersonCivilRegistrationIdentifier>1111111111</PersonCivilRegistrationIdentifier>
            <PersonGivenName>Bruce</PersonGivenName>
            <PersonSurnameName>Lee</PersonSurnameName>
            <Employment>
                <EmploymentIdentifier>11111</EmploymentIdentifier>
            </Employment>
        </Person>
        <Person>
            <PersonCivilRegistrationIdentifier>2222222222</PersonCivilRegistrationIdentifier>
            <PersonGivenName>Chuck</PersonGivenName>
            <PersonSurnameName>Norris</PersonSurnameName><Employment>
                <EmploymentIdentifier>22222</EmploymentIdentifier>
            </Employment>
        </Person>
    </GetPersonChangedAtDate20111201>
    """,
    parser=remove_blank_parser,
)

BRUCE_LEE = etree.XML(
    """
    <Person>
        <PersonCivilRegistrationIdentifier>1111111111</PersonCivilRegistrationIdentifier>
        <PersonGivenName>Bruce</PersonGivenName>
        <PersonSurnameName>Lee</PersonSurnameName>
        <Employment>
            <EmploymentIdentifier>11111</EmploymentIdentifier>
        </Employment>
    </Person>
    """,
    parser=remove_blank_parser,
)


class TestTarGzHelper:
    """
    Generate .tar.gz files to be used as test fixtures
    """

    def setup_class(self):
        self.FIXTURE_FOLDER = tempfile.TemporaryDirectory()

        self.tar_gz_file1 = Path(self.FIXTURE_FOLDER.name).joinpath(Path("tar1.tar.gz"))
        with tarfile.open(str(self.tar_gz_file1), "w:gz") as tar:
            tar.add("tests/fixtures/tar_gz1/opt", arcname="opt")

        self.tar_gz_file2 = Path(self.FIXTURE_FOLDER.name).joinpath(Path("tar2.tar.gz"))
        with tarfile.open(str(self.tar_gz_file2), "w:gz") as tar:
            tar.add("tests/fixtures/tar_gz2/opt", arcname="opt")

    def teardown_class(self):
        self.FIXTURE_FOLDER.cleanup()

    def get_xml_responses_from_tar_gz_file(self):
        log_file_lines = extract_log_file_lines(self.tar_gz_file1)
        xml_responses = get_sd_xml_responses(log_file_lines)
        return xml_responses


class TestGetTarGzArchiveFiles(TestTarGzHelper):
    def test_returns_tar_gz_files_in_folder(self):
        path = Path(self.FIXTURE_FOLDER.name)
        tar_gz_files = get_tar_gz_archive_files(path)

        assert len(tar_gz_files) == 2
        assert "tar1.tar.gz" == tar_gz_files[0].name
        assert "tar2.tar.gz" == tar_gz_files[1].name

    def test_returns_empty_list_when_no_tar_files_in_folder(self):
        path = Path("")
        tar_gz_files = get_tar_gz_archive_files(path)

        assert [] == tar_gz_files


class TestExtractLogFileLines(TestTarGzHelper):
    @patch("tarfile.TarFile.extractfile")
    def test_log_file_does_not_exist(self, mock_tarfile):
        mock_tarfile.side_effect = KeyError()

        assert [] == extract_log_file_lines(self.tar_gz_file1)


class TestGetSdXmlResponses(TestTarGzHelper):
    def test_log_file_should_contain_two_elements(self):
        print(self.tar_gz_file1)
        xml_responses = self.get_xml_responses_from_tar_gz_file()
        print(xml_responses)

        assert isinstance(xml_responses, list)
        assert len(xml_responses) == 2
        assert etree.iselement(xml_responses[0])
        assert etree.iselement(xml_responses[1])

    def test_return_empty_list_when_no_sd_xml_responses(self):
        log_file_lines = ["no", "sd", "responses"]
        xml_responses = get_sd_xml_responses(log_file_lines)

        assert [] == xml_responses


class TestGetSdPersonChangedAtDateResponses(TestTarGzHelper):
    def test_should_contain_a_single_person_changed_at_date_response(self):
        xml_responses = self.get_xml_responses_from_tar_gz_file()
        changed_at_date_responses = get_sd_person_changed_at_date_responses(
            xml_responses
        )

        assert isinstance(changed_at_date_responses, list)
        assert len(changed_at_date_responses) == 1

        assert etree.tostring(XML_FIXTURE) == etree.tostring(
            changed_at_date_responses[0]
        )


class TestGetPerson(TestTarGzHelper):
    @parameterized.expand([(IdType.CPR, "1111111111"), (IdType.EMPLOYMENT_ID, "11111")])
    def test_get_person_from_identifier(self, id_type, id_):
        actual_person = get_sd_person_changed(id_type, id_, XML_FIXTURE).change

        assert etree.tostring(BRUCE_LEE) == etree.tostring(actual_person)

    def test_return_none_when_no_persons_exists(self):
        xml_root = etree.XML(
            """
            <GetPersonChangedAtDate20111201>
                <RequestStructure>
                    <InstitutionIdentifier>XY</InstitutionIdentifier>
                    <ActivationDate>2021-09-14</ActivationDate>
                    <ActivationTime>00:00:00</ActivationTime>
                    <DeactivationDate>2021-09-15</DeactivationDate>
                    <DeactivationTime>00:00:59</DeactivationTime>
                    <ContactInformationIndicator>false</ContactInformationIndicator>
                    <PostalAddressIndicator>false</PostalAddressIndicator>
                </RequestStructure>
            </GetPersonChangedAtDate20111201>
            """,
            parser=remove_blank_parser,
        )

        actual = get_sd_person_changed(IdType.EMPLOYMENT_ID, "11111", xml_root).change

        assert actual is None

    def test_request_date_timestamps_set_correctly(self):
        sd_person_changed = get_sd_person_changed(
            IdType.EMPLOYMENT_ID, "11111", XML_FIXTURE
        )

        assert "2021-09-14" == sd_person_changed.start_date
        assert "2021-09-15" == sd_person_changed.end_date


class TestGetAllSdPersonChanges(TestTarGzHelper):
    def test_get_sd_person_changes_from_all_tar_gz_files(self):
        all_changes = get_all_sd_person_changes(
            IdType.EMPLOYMENT_ID, "11111", Path(self.FIXTURE_FOLDER.name)
        )

        assert "2021-09-14" == all_changes[0].start_date
        assert "2021-09-15" == all_changes[0].end_date
        assert etree.tostring(BRUCE_LEE) == etree.tostring(all_changes[0].change)

        # The same SD person is used in the fixture for convenience
        assert "2021-09-19" == all_changes[1].start_date
        assert "2021-09-20" == all_changes[1].end_date
        assert etree.tostring(BRUCE_LEE) == etree.tostring(all_changes[1].change)


class TestOutputToFile:
    @patch("builtins.open")
    @patch("lxml.etree.tostring")
    def test_skip_tars_with_no_changes(self, mock_etree, mock_open):
        all_changes = [
            SdPersonChange(
                start_date="not used here", end_date="not used here", change=BRUCE_LEE
            ),
            SdPersonChange(start_date="not used here", end_date="not used here"),
            SdPersonChange(
                start_date="not used here", end_date="not used here", change=BRUCE_LEE
            ),
        ]

        output_to_file(all_changes, Path("/tmp/dipex-writeable-file.txt"))

        assert 2 == mock_etree.call_count



from enum import Enum
from typing import Optional
from uuid import UUID

from pydantic import BaseModel


class JobFunction(str, Enum):
    job_position_identifier = "JobPositionIdentifier"
    employment_name = "EmploymentName"


# TODO: replace these models with the one present in the new SD client

# TODO: rename
class EmploymentWithTelephoneNumberIdentifier(BaseModel):
    employment_identifier: str
    telephone_number_ids: list[str] = []


class SDBasePerson(BaseModel):
    cpr: str
    given_name: Optional[str]
    surname: Optional[str]
    emp_with_telephone_number_identifiers: list[
        EmploymentWithTelephoneNumberIdentifier
    ] = []


class MOBasePerson(BaseModel):
    cpr: str
    givenname: str
    surname: str
    name: str
    uuid: UUID


class ITUserSystem(BaseModel):
    # UUID of the IT-system itself
    uuid: UUID
    # User key of the IT-user
    user_key: str



import logging
import uuid
from csv import DictWriter
from datetime import date
from datetime import datetime
from datetime import timedelta
from functools import partial
from io import StringIO
from itertools import chain
from itertools import starmap
from itertools import zip_longest
from operator import attrgetter
from operator import itemgetter
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional

import dar_helper
import los_files
import payloads as mo_payloads
import pydantic
import util
import uuids
from more_itertools import bucket
from more_itertools import first
from more_itertools import flatten
from more_itertools import last
from more_itertools import partition
from more_itertools import split_when
from more_itertools import unzip
from pydantic import Field
from pydantic import validator
from ra_utils.generate_uuid import uuid_generator

logger = logging.getLogger(__name__)


class OrgUnitBase(pydantic.BaseModel):
    start_date: date = Field(alias="StartDato")
    end_date: date = Field(alias="SlutDato")

    @validator("start_date", "end_date", pre=True)
    def validate_date(cls, v, values):
        if isinstance(v, date):
            return v

        formats = ["%d-%m-%Y", "%Y-%m-%d %H:%M:%S.%f"]
        for format in formats:
            try:
                parsed = datetime.strptime(v, format).date().isoformat()
                return parsed
            except ValueError:
                continue
        raise ValueError(f"Unable to parse date '{v}'")


class OrgUnit(OrgUnitBase):
    org_uuid: uuid.UUID = Field(alias="OrgUUID")
    bvn: str = Field(alias="BrugervendtNøgle")
    org_unit_name: Optional[str] = Field(alias="OrgEnhedsNavn")
    org_unit_type_uuid: Optional[uuid.UUID] = Field(alias="OrgEnhedsTypeUUID")
    parent_uuid: uuid.UUID = Field(alias="ParentUUID")
    is_in_line_org: bool = Field(default=False, alias="Med-i-LinjeOrg")
    los_id: Optional[str] = Field(alias="LOSID")
    cvr: Optional[str] = Field(alias="CVR")
    ean: Optional[str] = Field(alias="EAN")
    p_number: Optional[str] = Field(alias="P-Nr")
    se_number: Optional[str] = Field(alias="SE-Nr")
    int_debitor_number: Optional[str] = Field(alias="IntDebitor-Nr")
    mag_id: Optional[str] = Field(alias="MagID")
    post_address: Optional[str] = Field(default="", alias="PostAdresse")


class FailedDARLookup(pydantic.BaseModel):
    org_uuid: uuid.UUID
    post_address: str


class OrgUnitImporter:
    def __init__(self):
        self.dar_cache = {}
        self.uuid_generator = uuid_generator("AAK")

    def generate_unit_payload(self, orgunit: OrgUnit):
        from_date, to_date = util.convert_validities(
            orgunit.start_date, orgunit.end_date
        )
        parent_uuid = orgunit.parent_uuid
        if parent_uuid == orgunit.org_uuid:
            parent_uuid = uuids.ORG_UUID

        return mo_payloads.create_org_unit(
            uuid=orgunit.org_uuid,
            user_key=orgunit.bvn,
            name=orgunit.org_unit_name,
            parent_uuid=parent_uuid,
            org_unit_hierarchy=uuids.LINJE_ORG_HIERARCHY
            if orgunit.is_in_line_org
            else None,
            org_unit_type_uuid=orgunit.org_unit_type_uuid,
            from_date=from_date,
            to_date=to_date,
        )

    def generate_los_payload(self, orgunit: OrgUnit):
        from_date, to_date = util.convert_validities(
            orgunit.start_date, orgunit.end_date
        )
        obj_uuid = self.uuid_generator(str(orgunit.org_uuid) + "los")
        assert orgunit.los_id
        return mo_payloads.create_address(
            uuid=obj_uuid,
            value=orgunit.los_id,
            address_type_uuid=uuids.UNIT_LOS,
            org_unit_uuid=orgunit.org_uuid,
            from_date=from_date,
            to_date=to_date,
        )

    def generate_cvr_payload(self, orgunit: OrgUnit):
        from_date, to_date = util.convert_validities(
            orgunit.start_date, orgunit.end_date
        )
        obj_uuid = self.uuid_generator(str(orgunit.org_uuid) + "cvr")
        assert orgunit.cvr
        return mo_payloads.create_address(
            uuid=obj_uuid,
            value=orgunit.cvr,
            address_type_uuid=uuids.UNIT_CVR,
            org_unit_uuid=orgunit.org_uuid,
            from_date=from_date,
            to_date=to_date,
        )

    def generate_ean_payload(self, orgunit: OrgUnit):
        from_date, to_date = util.convert_validities(
            orgunit.start_date, orgunit.end_date
        )
        obj_uuid = self.uuid_generator(str(orgunit.org_uuid) + "ean")
        assert orgunit.ean
        return mo_payloads.create_address(
            uuid=obj_uuid,
            value=orgunit.ean,
            address_type_uuid=uuids.UNIT_EAN,
            org_unit_uuid=orgunit.org_uuid,
            from_date=from_date,
            to_date=to_date,
        )

    def generate_pnr_payload(self, orgunit: OrgUnit):
        from_date, to_date = util.convert_validities(
            orgunit.start_date, orgunit.end_date
        )
        obj_uuid = self.uuid_generator(str(orgunit.org_uuid) + "pnr")
        assert orgunit.p_number
        return mo_payloads.create_address(
            uuid=obj_uuid,
            value=orgunit.p_number,
            address_type_uuid=uuids.UNIT_PNR,
            org_unit_uuid=orgunit.org_uuid,
            from_date=from_date,
            to_date=to_date,
        )

    def generate_senr_payload(self, orgunit: OrgUnit):
        from_date, to_date = util.convert_validities(
            orgunit.start_date, orgunit.end_date
        )
        obj_uuid = self.uuid_generator(str(orgunit.org_uuid) + "senr")
        assert orgunit.se_number
        return mo_payloads.create_address(
            uuid=obj_uuid,
            value=orgunit.se_number,
            address_type_uuid=uuids.UNIT_SENR,
            org_unit_uuid=orgunit.org_uuid,
            from_date=from_date,
            to_date=to_date,
        )

    def generate_debitornr_payload(self, orgunit: OrgUnit):
        from_date, to_date = util.convert_validities(
            orgunit.start_date, orgunit.end_date
        )
        obj_uuid = self.uuid_generator(str(orgunit.org_uuid) + "debitornr")
        assert orgunit.int_debitor_number
        return mo_payloads.create_address(
            uuid=obj_uuid,
            value=orgunit.int_debitor_number,
            address_type_uuid=uuids.UNIT_DEBITORNR,
            org_unit_uuid=orgunit.org_uuid,
            from_date=from_date,
            to_date=to_date,
        )

    def generate_postaddr_payload(self, orgunit: OrgUnit):
        from_date, to_date = util.convert_validities(
            orgunit.start_date, orgunit.end_date
        )
        obj_uuid = self.uuid_generator(str(orgunit.org_uuid) + "postaddr")
        assert orgunit.post_address
        return mo_payloads.create_address(
            uuid=obj_uuid,
            value=self.dar_cache[orgunit.post_address],
            address_type_uuid=uuids.UNIT_POSTADDR,
            org_unit_uuid=orgunit.org_uuid,
            from_date=from_date,
            to_date=to_date,
        )

    def generate_mag_id_payload(self, orgunit: OrgUnit):
        from_date, to_date = util.convert_validities(
            orgunit.start_date, orgunit.end_date
        )
        obj_uuid = self.uuid_generator(str(orgunit.org_uuid) + "mag_id")
        assert orgunit.mag_id
        return mo_payloads.create_address(
            uuid=obj_uuid,
            value=orgunit.mag_id,
            address_type_uuid=uuids.UNIT_MAG_ID,
            org_unit_uuid=orgunit.org_uuid,
            from_date=from_date,
            to_date=to_date,
        )

    @staticmethod
    def consolidate_payloads(payload_list: List[dict]):
        """
        Given a list of sorted payloads, collapse identical payloads that are
        immediately consecutive, i.e. payloads with the same fields,
        where the end date of one corresponds to the start date of the next

        The function assumes that input does not contain overlapping timespans
        """

        def payloads_identical(p1, p2):
            check_keys = (p1.keys() | p2.keys()) - {"validity"}
            return all(p1.get(k) == p2.get(k) for k in check_keys)

        def is_not_consecutive(p1, p2):
            p1_to = datetime.strptime(p1["validity"]["to"], "%Y-%m-%d")
            p2_from = datetime.strptime(p2["validity"]["from"], "%Y-%m-%d")
            return not (
                payloads_identical(p1, p2) and p1_to + timedelta(days=1) == p2_from
            )

        def merge_objects(group):
            start, end = first(group), last(group)
            start["validity"]["to"] = end["validity"]["to"]
            return start

        if len(payload_list) < 2:
            return payload_list
        consecutive_groups = split_when(payload_list, is_not_consecutive)
        return list(map(merge_objects, consecutive_groups))

    def write_failed_addresses(
        self,
        failed: List[FailedDARLookup],
        filename: str,
        delimiter: str = "#",
    ):
        """Write failed addresses to an external file"""
        output_filename = f"failed_addr_{filename}"  # `filename` is `xxx.csv`

        # Write CSV to output buffer
        output = StringIO()
        writer = DictWriter(output, ["org_uuid", "post_address"], delimiter=delimiter)
        writer.writeheader()
        writer.writerows(f.dict() for f in failed)

        # AaK never actually gave us write permissions to the FTP server. However, we had to try to write to the FTP
        # server, as this was part of the assignment, despite the fact that it would always fail. We also had to write
        # the CSV file, which is handled with the FS fileset

        # Write CSV file to FTP folder
        ftp_fileset = los_files.get_fileset_implementation()
        try:
            result = ftp_fileset.write_file(
                output_filename, output, folder="DARfejlliste"
            )
        except Exception:
            logger.exception("Could not write %r to FTP", output_filename)
        else:
            logger.info("Wrote %r to FTP, result = %r", output_filename, result)

        # Write CSV file to MO queries folder
        fs_fileset = los_files.FSFileSet()
        try:
            result = fs_fileset.write_file(output_filename, output)
        except IOError:
            logger.exception("Could not write %r to queries folder", output_filename)
        else:
            logger.info(
                "Wrote %r to queries folder, result = %r", output_filename, result
            )

    async def handle_addresses(self, org_units, filename):
        addresses = map(attrgetter("post_address"), org_units)
        addresses = set(filter(lambda address: address is not None, addresses))

        if len(addresses) == 0:
            return

        address_lookups = await dar_helper.dar_datavask_multiple(addresses)

        # Split into two lists where lookup succeeded and failed
        success, failure = partition(lambda x: x[1] is None, address_lookups)

        success = dict(success)
        logger.debug("%d addresses found", len(success))
        self.dar_cache.update(success)

        failed_addresses = set(map(itemgetter(0), failure))
        logger.debug("%d addresses could not be found", len(failed_addresses))
        if failed_addresses:
            # Join `failed_addresses` with `org_units` on `post_address` to get
            # `org_uuid` for each failed address lookup.
            joined: List[FailedDARLookup] = [
                FailedDARLookup(
                    org_uuid=org_unit.org_uuid,
                    post_address=org_unit.post_address,
                )
                for org_unit in org_units
                if org_unit.post_address in failed_addresses
            ]
            self.write_failed_addresses(joined, filename)

    def create_unit_payloads(self, org_units) -> Iterable[dict]:
        return map(self.generate_unit_payload, org_units)

    async def create_detail_payloads(self, org_units) -> Iterable[dict]:
        """
        Generate all relevant detail payloads based on the org unit objects
        All rows are run through different sets of generators and filters
        """

        generators = [
            (self.generate_los_payload, lambda orgunit: orgunit.los_id),
            (self.generate_cvr_payload, lambda orgunit: orgunit.cvr),
            (self.generate_ean_payload, lambda orgunit: orgunit.ean),
            (self.generate_pnr_payload, lambda orgunit: orgunit.p_number),
            (self.generate_senr_payload, lambda orgunit: orgunit.se_number),
            (
                self.generate_debitornr_payload,
                lambda orgunit: orgunit.int_debitor_number,
            ),
            (self.generate_mag_id_payload, lambda orgunit: orgunit.mag_id),
            (
                self.generate_postaddr_payload,
                lambda orgunit: orgunit.post_address
                and orgunit.post_address in self.dar_cache,
            ),
        ]

        def run_generator(generator, filter_fn) -> Iterator:
            return map(generator, filter(filter_fn, org_units))

        return flatten(chain(starmap(run_generator, generators)))

    async def handle_initial(self, filename):
        """
        Handles reading the special 'initial' file

        The file contains org unit data, as well as data on the associated details

        The initial org unit file contains historic data, so a minimal set of
        create/edit payloads are created accordingly
        """
        org_units = los_files.read_csv(filename, OrgUnit)

        await self.handle_addresses(org_units, filename)

        unit_payloads = self.create_unit_payloads(org_units)
        detail_payloads = await self.create_detail_payloads(org_units)
        payloads = list(unit_payloads) + list(detail_payloads)

        # Bucket all payloads referring to the same object
        uuid_buckets = bucket(payloads, key=lambda payload: payload["uuid"])
        sorted_buckets = map(
            lambda uuid_key: sorted(
                uuid_buckets[uuid_key], key=lambda x: x["validity"]["from"]
            ),
            uuid_buckets,
        )
        consolidated_buckets = list(map(self.consolidate_payloads, sorted_buckets))

        split_lists = map(lambda x: (x[0], x[1:]), consolidated_buckets)
        heads, tails = unzip(split_lists)

        # OS2mo reads an object before performing an edit to it, so we need to ensure
        # that we don't perform multiple edits to an object in parallel, which could
        # cause one edit to be overwritten by another
        # We create layers containing at most one edit request for each org unit UUID,
        # and execute the layers sequentially, while allowing the importer to submit the
        # individual requests in a layer in parallel
        edit_payloads = map(partial(map, mo_payloads.convert_create_to_edit), tails)
        edit_layers = zip_longest(*edit_payloads)
        edit_layers_filtered = map(partial(filter, None.__ne__), edit_layers)

        async with util.get_client_session() as session:
            await util.create_details(session, heads)
            for edit_layer in edit_layers_filtered:
                await util.edit_details(session, edit_layer)

    async def handle_create(self, filename):
        """
        Handle creating new org units and details
        We are guaranteed to only have one row per org unit
        """
        org_units = los_files.read_csv(filename, OrgUnit)

        await self.handle_addresses(org_units, filename)

        org_unit_payloads = self.create_unit_payloads(org_units)
        detail_payloads = await self.create_detail_payloads(org_units)

        async with util.get_client_session() as session:
            await util.create_details(
                session, chain(org_unit_payloads, detail_payloads)
            )

    async def handle_edit(self, filename: str, filedate: datetime):
        """
        Handle changes to existing org units and details
        We are guaranteed to only have one row per org unit

        New details on an existing org unit will show up in this file, rather than the
        'nye' file. So we have to potentially perform inserts of new data.

        As a row contains information about the org unit as well as its details,
        we do not know what has been changed. However, all information is managed
        by the external system so we can safely reimport the "same" data, as opposed to
        trying to compare the existing objects in OS2mo
        """
        org_units = los_files.read_csv(filename, OrgUnit)
        org_unit_payloads = self.create_unit_payloads(org_units)
        detail_payloads = await self.create_detail_payloads(org_units)

        orgfunk_uuids = set(await util.lookup_organisationfunktion())
        detail_creates, detail_edits = partition(
            lambda payload: payload["uuid"] in orgfunk_uuids, detail_payloads
        )
        converter = partial(
            mo_payloads.convert_create_to_edit, from_date=filedate.date().isoformat()
        )
        edits = map(converter, chain(org_unit_payloads, detail_edits))

        async with util.get_client_session() as session:
            await util.create_details(session, detail_creates)
            await util.edit_details(session, edits)

    async def run(self, last_import: datetime):
        """
        Reads org unit files newer than last_import
        and performs inserts/updates as needed
        """
        logger.info("Starting org unit import")
        filenames = los_files.get_fileset_implementation().get_import_filenames()

        initials = los_files.parse_filenames(
            filenames, prefix="Org_inital", last_import=last_import
        )
        creates = los_files.parse_filenames(
            filenames, prefix="Org_nye", last_import=last_import
        )
        edits = los_files.parse_filenames(
            filenames, prefix="Org_ret", last_import=last_import
        )

        for filename, _ in initials:
            await self.handle_initial(filename)

        for filename, _ in creates:
            await self.handle_create(filename)

        for filename, filedate in edits:
            await self.handle_edit(filename, filedate)

        logger.info("Org unit import done")



from datetime import date
from unittest.mock import MagicMock, patch
from uuid import UUID

from sdlon.models import ITUserSystem
from tests.fixtures import mock_graphql_client

from sdlon.it_systems import (
    get_sd_to_ad_it_system_uuid,
    QUERY_GET_SD_TO_AD_IT_SYSTEM_UUID,
    get_employee_it_systems,
    QUERY_GET_EMPLOYEE_IT_SYSTEMS,
    add_it_system_to_employee,
    MUTATION_ADD_IT_SYSTEM_TO_EMPLOYEE,
)


def test_get_sd_to_ad_it_system_uuid(mock_graphql_client: MagicMock) -> None:
    # Arrange
    mock_execute = MagicMock(
        return_value={
            "itsystems": {"objects": [{"uuid": "988dead8-7564-464a-8339-b7057bfa2665"}]}
        }
    )
    mock_graphql_client.execute = mock_execute

    # Act
    it_system_uuid = get_sd_to_ad_it_system_uuid(
        mock_graphql_client, "AD-bruger til SD"
    )

    # Assert
    mock_execute.assert_called_once_with(
        QUERY_GET_SD_TO_AD_IT_SYSTEM_UUID,
        variable_values={"user_key": "AD-bruger til SD"},
    )
    assert it_system_uuid == UUID("988dead8-7564-464a-8339-b7057bfa2665")


def test_get_employee_it_systems(mock_graphql_client: MagicMock) -> None:
    # Arrange
    mock_execute = MagicMock(
        return_value={
            "employees": {
                "objects": [
                    {
                        "current": {
                            "itusers": [
                                {
                                    "itsystem": {
                                        "uuid": "5168dd45-4cb5-4932-b8a1-10dbe736fc5d"
                                    },
                                    "user_key": "user_key1",
                                },
                                {
                                    "itsystem": {
                                        "uuid": "a1608e69-c422-404f-a6cc-b873c50af111"
                                    },
                                    "user_key": "user_key2",
                                },
                            ]
                        }
                    }
                ]
            }
        }
    )
    mock_graphql_client.execute = mock_execute

    # Act
    it_systems = get_employee_it_systems(
        mock_graphql_client, UUID("353ed4ae-f489-11ed-bba5-0bfbad9d10d2")
    )

    # Assert
    mock_execute.assert_called_once_with(
        QUERY_GET_EMPLOYEE_IT_SYSTEMS,
        variable_values={"uuid": "353ed4ae-f489-11ed-bba5-0bfbad9d10d2"},
    )
    assert it_systems == [
        ITUserSystem(
            uuid=UUID("5168dd45-4cb5-4932-b8a1-10dbe736fc5d"),
            user_key="user_key1",
        ),
        ITUserSystem(
            uuid=UUID("a1608e69-c422-404f-a6cc-b873c50af111"),
            user_key="user_key2",
        ),
    ]


@patch("sdlon.it_systems.date")
def test_add_it_system_to_employee(
    mock_date: MagicMock,
    mock_graphql_client: MagicMock,
) -> None:
    # Arrange
    mock_execute = MagicMock()
    mock_graphql_client.execute = mock_execute
    mock_date.today = MagicMock(return_value=date(2000, 1, 1))

    # Act
    add_it_system_to_employee(
        mock_graphql_client,
        UUID("cfd0b8d2-f48b-11ed-a440-37f1efd1415a"),
        UUID("d0c009f0-f48b-11ed-ba89-ff38216510fc"),
        "AD-bruger fra SD",
    )

    # Assert
    mock_execute.assert_called_once_with(
        MUTATION_ADD_IT_SYSTEM_TO_EMPLOYEE,
        variable_values={
            "input": {
                "user_key": "AD-bruger fra SD",
                "itsystem": "d0c009f0-f48b-11ed-ba89-ff38216510fc",
                "validity": {"from": "2000-01-01"},
                "person": "cfd0b8d2-f48b-11ed-a440-37f1efd1415a",
            }
        },
    )



import unittest.mock
from collections import OrderedDict
from copy import deepcopy

import pytest
from more_itertools import one

from .fixtures import get_read_employment_changed_fixture
from sdlon.engagement import _is_external, filtered_professions
from sdlon.engagement import (
    create_engagement,
    is_employment_id_and_no_salary_minimum_consistent,
)


class TestIsExternal(unittest.TestCase):
    def test_is_external_is_false_for_number_employment_id(self):
        assert not _is_external("12345")
        assert not _is_external("1")

    def test_is_external_is_true_for_non_number_employment_id(self):
        assert _is_external("ABCDE")


class TestIsEmploymentIdAndNoSalaryMinimumConsistent:
    @pytest.mark.parametrize(
        "no_salary_minimum,employment_id,job_pos_id,expected",
        [
            (None, "12345", 1, True),
            (9000, "External", 10000, True),
            (9000, "External", 8000, False),
            (9000, "12345", 8000, True),
            (9000, "12345", 10000, False),
        ],
    )
    def test_return_values(
        self, no_salary_minimum, employment_id, job_pos_id, expected
    ):
        engagement = one(
            get_read_employment_changed_fixture(
                employment_id=employment_id,
                job_pos_id=job_pos_id,
            )
        )["Employment"]

        assert (
            is_employment_id_and_no_salary_minimum_consistent(
                engagement, no_salary_minimum
            )
            == expected
        )

    @pytest.mark.parametrize(
        "job_pos_id2,expected",
        [
            ("1001", True),
            ("9001", False),
        ],
    )
    def test_job_pos_ids_consistent_but_different(self, job_pos_id2, expected):
        # Arrange
        engagement = one(get_read_employment_changed_fixture())["Employment"]

        profession1 = deepcopy(engagement.get("Profession"))
        profession2 = deepcopy(engagement.get("Profession"))
        profession2["JobPositionIdentifier"] = job_pos_id2
        engagement.update({"Profession": [profession1, profession2]})

        # Assert
        assert (
            is_employment_id_and_no_salary_minimum_consistent(engagement, 9000)
            == expected
        )


class TestCreateEngagement(unittest.TestCase):
    @unittest.mock.patch("sdlon.engagement.read_employment_at")
    def test_return_none_when_sd_employment_not_found(self, mock_read_employment_at):
        mock_read_employment_at.return_value = None
        mock_sd_updater = unittest.mock.MagicMock()
        create_engagement(mock_sd_updater, 12345, "person_uuid")

        mock_sd_updater.assert_not_called()


def test_filter_multiple_professions():
    # Arrange
    sd_employment = OrderedDict(
        {
            "EmploymentIdentifier": "12345",
            "Profession": [
                {"JobPositionIdentifier": "1"},
                {"JobPositionIdentifier": "2"},
                {"JobPositionIdentifier": "3"},
                {"JobPositionIdentifier": "4"},
                {"JobPositionIdentifier": "5"},
            ],
        }
    )

    # Act
    filtered_employment = filtered_professions(sd_employment, ["1", "2"])

    # Assert
    assert filtered_employment == OrderedDict(
        {
            "EmploymentIdentifier": "12345",
            "Profession": [
                {"JobPositionIdentifier": "3"},
                {"JobPositionIdentifier": "4"},
                {"JobPositionIdentifier": "5"},
            ],
        }
    )


def test_filter_single_professions():
    # Arrange
    sd_employment = OrderedDict(
        {
            "EmploymentIdentifier": "12345",
            "Profession": {"JobPositionIdentifier": "3"},
        }
    )

    # Act
    filtered_employment = filtered_professions(sd_employment, ["1", "2"])

    # Assert
    assert filtered_employment == OrderedDict(
        {
            "EmploymentIdentifier": "12345",
            "Profession": [{"JobPositionIdentifier": "3"}],
        }
    )



import datetime
import logging
import typing
from typing import Tuple

import click
import ra_utils.ensure_single_run
from alembic.migration import MigrationContext
from alembic.operations import Operations
from more_itertools import ichunked
from ra_utils.ensure_single_run import ensure_single_run
from ra_utils.job_settings import JobSettings
from ra_utils.load_settings import load_settings
from ra_utils.tqdm_wrapper import tqdm
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker

from .gql_lora_cache_async import GQLLoraCache
from .lora_cache import get_cache as LoraCache
from .old_lora_cache import OldLoraCache
from .sql_table_defs import Adresse
from .sql_table_defs import Base
from .sql_table_defs import Bruger
from .sql_table_defs import DARAdresse
from .sql_table_defs import Engagement
from .sql_table_defs import Enhed
from .sql_table_defs import Enhedssammenkobling
from .sql_table_defs import Facet
from .sql_table_defs import ItForbindelse
from .sql_table_defs import ItSystem
from .sql_table_defs import Klasse
from .sql_table_defs import KLE
from .sql_table_defs import Kvittering
from .sql_table_defs import Leder
from .sql_table_defs import LederAnsvar
from .sql_table_defs import Orlov
from .sql_table_defs import Rolle
from .sql_table_defs import Tilknytning
from .sql_url import DatabaseFunction
from .sql_url import generate_connection_url
from .sql_url import generate_engine_settings


class SqlExportSettings(JobSettings):
    class Config:
        settings_json_prefix = "exporters.actual_state"


logger = logging.getLogger(__name__)


class SqlExport:
    def __init__(self, force_sqlite=False, historic=False, settings=None):
        logger.info("Start SQL export")
        self.force_sqlite = force_sqlite
        self.historic = historic
        self.settings = settings
        self.engine = self._get_engine()
        self.export_cpr = self._get_export_cpr_setting()
        self.chunk_size = 5000

    def _get_engine(self) -> Engine:
        database_function = DatabaseFunction.ACTUAL_STATE
        if self.historic:
            database_function = DatabaseFunction.ACTUAL_STATE_HISTORIC
        db_string = generate_connection_url(
            database_function, force_sqlite=self.force_sqlite, settings=self.settings
        )
        engine_settings = generate_engine_settings(
            database_function, force_sqlite=self.force_sqlite, settings=self.settings
        )
        return create_engine(db_string, **engine_settings)

    def _get_export_cpr_setting(self) -> bool:
        return self.settings.get("exporters.actual_state.export_cpr", True)

    def _get_lora_cache(
        self, resolve_dar, use_pickle
    ) -> typing.Union[OldLoraCache, GQLLoraCache]:
        if self.historic:
            lc = LoraCache(
                resolve_dar=resolve_dar, full_history=True, settings=self.settings
            )
            lc.populate_cache(dry_run=use_pickle)
        else:
            lc = LoraCache(resolve_dar=resolve_dar, settings=self.settings)
            lc.populate_cache(dry_run=use_pickle)
            lc.calculate_derived_unit_data()
            lc.calculate_primary_engagements()
        return lc

    def _get_db_session(self) -> Session:
        Session = sessionmaker(bind=self.engine, autoflush=False)
        return Session()

    def _get_lora_class(self, uuid: str) -> Tuple[str, dict]:
        cls: dict = self.lc.classes.get(uuid) or {"title": uuid}
        return uuid, cls

    def perform_export(self, resolve_dar=True, use_pickle=None):
        def timestamp():
            return datetime.datetime.now()

        trunc_tables = dict(Base.metadata.tables)
        trunc_tables.pop("kvittering")

        logger.info("Dropping tables")
        Base.metadata.drop_all(self.engine, tables=trunc_tables.values())
        logger.info("Creating tables")
        Base.metadata.create_all(self.engine)

        self.session = self._get_db_session()

        query_time = timestamp()
        kvittering = self._add_receipt(query_time)
        self.lc = self._get_lora_cache(resolve_dar, use_pickle)

        start_delivery_time = timestamp()
        self._update_receipt(kvittering, start_delivery_time)

        tasks = [
            self._add_classification,
            self._add_users_and_units,
            self._add_addresses,
            self._add_dar_addresses,
            self._add_engagements,
            self._add_associations,
            self._add_managers,
            self._add_it_systems,
            self._add_kles,
            self._add_related,
        ]
        for task in tqdm(tasks, desc="SQLExport", unit="task"):
            task()

        end_delivery_time = timestamp()
        self._update_receipt(kvittering, start_delivery_time, end_delivery_time)

    def get_actual_tables(self):
        connection = self.engine.connect()
        inspector = Inspector.from_engine(connection)
        actual_tables = inspector.get_table_names()
        return set(actual_tables)

    def swap_tables(self):
        """Swap tables around to present the exported data.

        Swaps the current tables to old tables, then swaps write tables to current.
        Finally drops the old tables leaving just the current tables.
        """
        logger.info("Swapping tables")
        connection = self.engine.connect()
        ctx = MigrationContext.configure(connection)
        op = Operations(ctx)

        def gen_table_names(write_table):
            """Generate current and old table names from write tables."""
            # Current tables do not have the prefix 'w'
            current_table = write_table[1:]
            old_table = current_table + "_old"
            return write_table, current_table, old_table

        tables = dict(Base.metadata.tables)
        tables.pop("kvittering")
        tables = tables.keys()
        tables = list(map(gen_table_names, tables))

        # Drop any left-over old tables that may exist
        with ctx.begin_transaction():
            actual_tables = self.get_actual_tables()
            for _, _, old_table in tables:
                if old_table in actual_tables:
                    op.drop_table(old_table)

        # Rename current to old and write to current
        with ctx.begin_transaction():
            actual_tables = self.get_actual_tables()
            for write_table, current_table, old_table in tables:
                if current_table in actual_tables:
                    op.rename_table(current_table, old_table)
                # Rename write table to current table
                op.rename_table(write_table, current_table)

        # Drop any old tables that may exist
        with ctx.begin_transaction():
            actual_tables = self.get_actual_tables()
            for _, _, old_table in tables:
                if old_table in actual_tables:
                    op.drop_table(old_table)

    def _add_classification(self, output=False):
        logger.info("Add classification")
        facets = tqdm(self.lc.facets.items(), desc="Export facet", unit="facet")
        for chunk in ichunked(facets, self.chunk_size):
            for facet, facet_info in chunk:
                sql_facet = Facet(
                    uuid=facet,
                    bvn=facet_info["user_key"],
                )
                self.session.add(sql_facet)
            self.session.commit()

        classes = tqdm(self.lc.classes.items(), desc="Export class", unit="class")
        for chunk in ichunked(classes, self.chunk_size):
            for klasse, klasse_info in chunk:
                sql_class = Klasse(
                    uuid=klasse,
                    bvn=klasse_info["user_key"],
                    titel=klasse_info["title"],
                    facet_uuid=klasse_info["facet"],
                    facet_bvn=self.lc.facets[klasse_info["facet"]]["user_key"],
                )
                self.session.add(sql_class)
            self.session.commit()

        if output:
            for result in self.engine.execute("select * from facetter limit 4"):
                print(result.items())
            for result in self.engine.execute("select * from klasser limit 4"):
                print(result.items())

    def _add_users_and_units(self, output=False):
        logger.info("Add users and units")
        users = tqdm(self.lc.users.items(), desc="Export user", unit="user")
        for chunk in ichunked(users, self.chunk_size):
            for user, user_effects in chunk:
                for user_info in user_effects:
                    sql_user = Bruger(
                        uuid=user,
                        bvn=user_info["user_key"],
                        fornavn=user_info["fornavn"],
                        efternavn=user_info["efternavn"],
                        kaldenavn_fornavn=user_info["kaldenavn_fornavn"],
                        kaldenavn_efternavn=user_info["kaldenavn_efternavn"],
                        cpr=user_info["cpr"] if self.export_cpr else "",
                        startdato=user_info["from_date"],
                        slutdato=user_info["to_date"],
                    )
                    self.session.add(sql_user)
            self.session.commit()

        units = tqdm(self.lc.units.items(), desc="Export unit", unit="unit")
        for chunk in ichunked(units, self.chunk_size):
            for unit, unit_validities in chunk:
                for unit_info in unit_validities:
                    location = unit_info.get("location")
                    manager_uuid = unit_info.get("manager_uuid")
                    acting_manager_uuid = unit_info.get("acting_manager_uuid")

                    unit_type = unit_info["unit_type"]

                    enhedsniveau_titel = ""
                    if unit_info["level"]:
                        enhedsniveau_titel = self.lc.classes[unit_info["level"]][
                            "title"
                        ]

                    (
                        org_unit_hierarchy_uuid,
                        org_unit_hierarchy_class,
                    ) = self._get_lora_class(unit_info["org_unit_hierarchy"])

                    sql_unit = Enhed(
                        uuid=unit,
                        navn=unit_info["name"],
                        bvn=unit_info["user_key"],
                        forældreenhed_uuid=unit_info["parent"],
                        enhedstype_uuid=unit_type,
                        enhedsniveau_uuid=unit_info["level"],
                        organisatorisk_sti=location,
                        leder_uuid=manager_uuid,
                        fungerende_leder_uuid=acting_manager_uuid,
                        enhedstype_titel=self.lc.classes[unit_type]["title"],
                        enhedsniveau_titel=enhedsniveau_titel,
                        opmærkning_uuid=org_unit_hierarchy_uuid,
                        opmærkning_titel=org_unit_hierarchy_class["title"],
                        startdato=unit_info["from_date"],
                        slutdato=unit_info["to_date"],
                    )
                    self.session.add(sql_unit)
            self.session.commit()

        if output:
            for result in self.engine.execute("select * from brugere limit 5"):
                print(result)
            for result in self.engine.execute("select * from enheder limit 5"):
                print(result)

    def _add_engagements(self, output=False):
        logger.info("Add engagements")
        engagements = tqdm(
            self.lc.engagements.items(), desc="Export engagement", unit="engagement"
        )
        for chunk in ichunked(engagements, self.chunk_size):
            for engagement, engagement_validity in chunk:
                for engagement_info in engagement_validity:
                    if engagement_info["primary_type"] is not None:
                        primærtype_titel = self.lc.classes[
                            engagement_info["primary_type"]
                        ]["title"]
                    else:
                        primærtype_titel = ""

                    engagement_type_uuid = engagement_info["engagement_type"]
                    job_function_uuid, job_function_class = self._get_lora_class(
                        engagement_info["job_function"]
                    )

                    sql_engagement = Engagement(
                        uuid=engagement,
                        enhed_uuid=engagement_info["unit"],
                        bruger_uuid=engagement_info["user"],
                        bvn=engagement_info["user_key"],
                        engagementstype_uuid=engagement_info["engagement_type"],
                        primær_boolean=engagement_info.get("primary_boolean"),
                        arbejdstidsfraktion=engagement_info["fraction"],
                        engagementstype_titel=self.lc.classes.get(
                            engagement_type_uuid, {"title": engagement_type_uuid}
                        )["title"],
                        primærtype_titel=primærtype_titel,
                        stillingsbetegnelse_uuid=job_function_uuid,
                        stillingsbetegnelse_titel=job_function_class["title"],
                        primærtype_uuid=engagement_info["primary_type"],
                        startdato=engagement_info["from_date"],
                        slutdato=engagement_info["to_date"],
                        **engagement_info["extensions"],
                    )
                    self.session.add(sql_engagement)
            self.session.commit()

        if output:
            for result in self.engine.execute("select * from engagementer limit 5"):
                print(result.items())

    def _add_addresses(self, output=False):
        logger.info("Add addresses")
        addresses = tqdm(
            self.lc.addresses.items(), desc="Export address", unit="address"
        )
        for chunk in ichunked(addresses, self.chunk_size):
            for address, address_validities in chunk:
                for address_info in address_validities:
                    visibility_text = None
                    if address_info["visibility"] is not None:
                        visibility_text = self.lc.classes[address_info["visibility"]][
                            "title"
                        ]
                    visibility_scope = None
                    if address_info["visibility"] is not None:
                        visibility_scope = self.lc.classes[address_info["visibility"]][
                            "scope"
                        ]

                    sql_address = Adresse(
                        uuid=address,
                        enhed_uuid=address_info["unit"],
                        bruger_uuid=address_info["user"],
                        værdi=address_info["value"],
                        dar_uuid=address_info["dar_uuid"],
                        adressetype_uuid=address_info["adresse_type"],
                        adressetype_bvn=self.lc.classes[address_info["adresse_type"]][
                            "user_key"
                        ],
                        adressetype_scope=address_info["scope"],
                        adressetype_titel=self.lc.classes[address_info["adresse_type"]][
                            "title"
                        ],
                        synlighed_uuid=address_info["visibility"],
                        synlighed_scope=visibility_scope,
                        synlighed_titel=visibility_text,
                        startdato=address_info["from_date"],
                        slutdato=address_info["to_date"],
                    )
                    self.session.add(sql_address)
            self.session.commit()
        if output:
            for result in self.engine.execute("select * from adresser limit 10"):
                print(result.items())

    def _add_dar_addresses(self, output=False):
        logger.info("Add DAR addresses")
        dar = tqdm(self.lc.dar_cache.items(), desc="Export DAR", unit="DAR")
        for chunk in ichunked(dar, self.chunk_size):

            for address, address_info in chunk:
                sql_address = DARAdresse(
                    uuid=address,
                    **{
                        key: value
                        for key, value in address_info.items()
                        if key in DARAdresse.__table__.columns.keys() and key != "id"
                    },
                )
                self.session.add(sql_address)
            self.session.commit()
        if output:
            for result in self.engine.execute("select * from dar_adresser limit 10"):
                print(result.items())

    def _add_associations(self, output=False):
        logger.info("Add associations")
        associations = tqdm(
            self.lc.associations.items(), desc="Export association", unit="association"
        )
        for chunk in ichunked(associations, self.chunk_size):

            for association, association_validity in chunk:
                for association_info in association_validity:
                    (
                        association_type_uuid,
                        association_type_class,
                    ) = self._get_lora_class(association_info["association_type"])
                    job_function_uuid, job_function_class = self._get_lora_class(
                        association_info["job_function"]
                    )
                    sql_association = Tilknytning(
                        uuid=association,
                        bruger_uuid=association_info["user"],
                        enhed_uuid=association_info["unit"],
                        bvn=association_info["user_key"],
                        tilknytningstype_uuid=association_type_uuid,
                        tilknytningstype_titel=association_type_class["title"],
                        startdato=association_info["from_date"],
                        slutdato=association_info["to_date"],
                        it_forbindelse_uuid=association_info["it_user"],
                        stillingsbetegnelse_uuid=job_function_uuid,
                        stillingsbetegnelse_titel=job_function_class["title"],
                        primær_boolean=association_info.get("primary_boolean"),
                    )
                    self.session.add(sql_association)
            self.session.commit()
        roles = tqdm(self.lc.roles.items(), desc="Export role", unit="role")
        for chunk in ichunked(roles, self.chunk_size):

            for role, role_validity in chunk:
                for role_info in role_validity:
                    sql_role = Rolle(
                        uuid=role,
                        bruger_uuid=role_info["user"],
                        enhed_uuid=role_info["unit"],
                        rolletype_uuid=role_info["role_type"],
                        rolletype_titel=self.lc.classes[role_info["role_type"]][
                            "title"
                        ],
                        startdato=role_info["from_date"],
                        slutdato=role_info["to_date"],
                    )
                    self.session.add(sql_role)
            self.session.commit()

        leaves = tqdm(self.lc.leaves.items(), desc="Export leave", unit="leave")
        for chunk in ichunked(leaves, self.chunk_size):

            for leave, leave_validity in chunk:
                for leave_info in leave_validity:
                    leave_type = leave_info["leave_type"]
                    sql_leave = Orlov(
                        uuid=leave,
                        bvn=leave_info["user_key"],
                        bruger_uuid=leave_info["user"],
                        orlovstype_uuid=leave_type,
                        orlovstype_titel=self.lc.classes[leave_type]["title"],
                        engagement_uuid=leave_info["engagement"],
                        startdato=leave_info["from_date"],
                        slutdato=leave_info["to_date"],
                    )
                    self.session.add(sql_leave)
            self.session.commit()
        if output:
            for result in self.engine.execute("select * from tilknytninger limit 4"):
                print(result.items())
            for result in self.engine.execute("select * from orlover limit 4"):
                print(result.items())
            for result in self.engine.execute("select * from roller limit 4"):
                print(result.items())

    def _add_it_systems(self, output=False):
        logger.info("Add IT systems")
        itsystems = tqdm(
            self.lc.itsystems.items(), desc="Export itsystem", unit="itsystem"
        )
        for chunk in ichunked(itsystems, self.chunk_size):
            for itsystem, itsystem_info in chunk:
                sql_itsystem = ItSystem(uuid=itsystem, navn=itsystem_info["name"])
                self.session.add(sql_itsystem)
            self.session.commit()

        it_connections = tqdm(
            self.lc.it_connections.items(),
            desc="Export it connection",
            unit="it connection",
        )
        for chunk in ichunked(it_connections, self.chunk_size):

            for it_connection, it_connection_validity in chunk:
                for it_connection_info in it_connection_validity:
                    sql_it_connection = ItForbindelse(
                        uuid=it_connection,
                        it_system_uuid=it_connection_info["itsystem"],
                        bruger_uuid=it_connection_info["user"],
                        enhed_uuid=it_connection_info["unit"],
                        brugernavn=it_connection_info["username"],
                        startdato=it_connection_info["from_date"],
                        slutdato=it_connection_info["to_date"],
                        primær_boolean=it_connection_info.get("primary_boolean"),
                    )
                    self.session.add(sql_it_connection)
            self.session.commit()
        if output:
            for result in self.engine.execute("select * from it_systemer limit 2"):
                print(result.items())

            for result in self.engine.execute("select * from it_forbindelser limit 2"):
                print(result.items())

    def _add_kles(self, output=False):
        logger.info("Add KLES")
        kles = tqdm(self.lc.kles.items(), desc="Export KLE", unit="KLE")
        for chunk in ichunked(kles, self.chunk_size):

            for kle, kle_validity in chunk:
                for kle_info in kle_validity:
                    sql_kle = KLE(
                        uuid=kle,
                        enhed_uuid=kle_info["unit"],
                        kle_aspekt_uuid=kle_info["kle_aspect"],
                        kle_aspekt_titel=self.lc.classes[kle_info["kle_aspect"]][
                            "title"
                        ],
                        kle_nummer_uuid=kle_info["kle_number"],
                        kle_nummer_titel=self.lc.classes[kle_info["kle_number"]][
                            "title"
                        ],
                        startdato=kle_info["from_date"],
                        slutdato=kle_info["to_date"],
                    )
                    self.session.add(sql_kle)
            self.session.commit()
        if output:
            for result in self.engine.execute("select * from kle limit 10"):
                print(result.items())

    def _add_receipt(self, query_time, start_time=None, end_time=None, output=False):
        logger.info("Add Receipt")
        sql_kvittering = Kvittering(
            query_tid=query_time,
            start_levering_tid=start_time,
            slut_levering_tid=end_time,
        )
        self.session.add(sql_kvittering)
        self.session.commit()
        if output:
            for result in self.engine.execute("select * from kvittering limit 10"):
                print(result.items())
        return sql_kvittering

    def _update_receipt(
        self, sql_kvittering, start_time=None, end_time=None, output=False
    ):
        logger.info("Update Receipt")
        sql_kvittering.start_levering_tid = start_time
        sql_kvittering.slut_levering_tid = end_time
        self.session.commit()
        if output:
            for result in self.engine.execute("select * from kvittering limit 10"):
                print(result.items())

    def _add_related(self, output=False):
        logger.info("Add Enhedssammenkobling")
        relateds = tqdm(self.lc.related.items(), desc="Export related", unit="related")
        for chunk in ichunked(relateds, self.chunk_size):

            for related, related_validity in chunk:
                for related_info in related_validity:
                    sql_related = Enhedssammenkobling(
                        uuid=related,
                        enhed1_uuid=related_info["unit1_uuid"],
                        enhed2_uuid=related_info["unit2_uuid"],
                        startdato=related_info["from_date"],
                        slutdato=related_info["to_date"],
                    )
                    self.session.add(sql_related)
            self.session.commit()
        if output:
            for result in self.engine.execute(
                "select * from enhedssammenkobling limit 10"
            ):
                print(result.items())

    def _add_managers(self, output=False):
        logger.info("Add managers")
        managers = tqdm(self.lc.managers.items(), desc="Export manager", unit="manager")
        for chunk in ichunked(managers, self.chunk_size):

            for manager, manager_validity in chunk:
                for manager_info in manager_validity:
                    sql_manager = Leder(
                        uuid=manager,
                        bruger_uuid=manager_info["user"],
                        enhed_uuid=manager_info["unit"],
                        niveautype_uuid=manager_info["manager_level"],
                        ledertype_uuid=manager_info["manager_type"],
                        niveautype_titel=self.lc.classes[manager_info["manager_level"]][
                            "title"
                        ],
                        ledertype_titel=self.lc.classes[manager_info["manager_type"]][
                            "title"
                        ],
                        startdato=manager_info["from_date"],
                        slutdato=manager_info["to_date"],
                    )
                    self.session.add(sql_manager)

                    for responsibility in manager_info["manager_responsibility"]:
                        sql_responsibility = LederAnsvar(
                            leder_uuid=manager,
                            lederansvar_uuid=responsibility,
                            lederansvar_titel=self.lc.classes[responsibility]["title"],
                            startdato=manager_info["from_date"],
                            slutdato=manager_info["to_date"],
                        )
                        self.session.add(sql_responsibility)
            self.session.commit()
        if output:
            for result in self.engine.execute("select * from ledere limit 10"):
                print(result.items())
            for result in self.engine.execute("select * from leder_ansvar limit 10"):
                print(result.items())

    def log_overlapping_runs_aak(self):
        self.engine.execute(
            "INSERT INTO [dbo].[kvittering_afvigelse] "
            "([query_tid],[aarsag]) VALUES (getdate(), "
            "'Time-export: hopper over da foregående "
            "loop stadig kører.')"
        )

    def export(self, resolve_dar: bool, use_pickle: typing.Any) -> None:

        self.perform_export(
            resolve_dar=resolve_dar,
            use_pickle=use_pickle,
        )

        self.swap_tables()


def wrap_export(args: dict, settings: dict) -> None:

    sql_export = SqlExport(
        force_sqlite=args["force_sqlite"],
        historic=args["historic"],
        settings=settings,
    )
    try:
        lock_name = "sql_export_actual"

        if args["historic"]:
            lock_name = "sql_export_historic"

        ensure_single_run(
            func=sql_export.export,
            lock_name=lock_name,
            resolve_dar=args["resolve_dar"],
            use_pickle=args["read_from_cache"],
        )

    except ra_utils.ensure_single_run.LockTaken as name_of_lock:
        logger.warning(f"Lock {name_of_lock} taken, aborting export")
        if "log_overlapping_aak" in settings and settings.get("log_overlapping_aak"):
            sql_export.log_overlapping_runs_aak()


@click.command(help="SQL export")
@click.option("--resolve-dar", is_flag=True, envvar="RESOLVE_DAR")
@click.option("--historic", is_flag=True)
@click.option("--read-from-cache", is_flag=True, envvar="USE_CACHED_LORACACHE")
@click.option("--force-sqlite", is_flag=True)
def cli(**args):
    """
    Command line interface.
    """
    pydantic_settings = SqlExportSettings()
    pydantic_settings.start_logging_based_on_settings()
    logger.info("Command line args: %r", args)

    settings = load_settings()

    wrap_export(args=args, settings=settings)

    logger.info("*SQL export ended*")


if __name__ == "__main__":
    cli()



from io import BytesIO
from io import StringIO
from unittest import mock

from hypothesis import given
from hypothesis import settings
from hypothesis import strategies as st

from .helpers import mock_config


class TestFTPFileSet:
    @given(st.text())
    def test_convert_stringio_to_bytesio(self, text):
        fileset = self._get_ftpfileset()
        stringio = StringIO(text)
        bytesio = fileset._convert_stringio_to_bytesio(stringio)
        assert isinstance(bytesio, BytesIO)

    @settings(deadline=None)
    @given(st.text(), st.text(), st.text())
    def test_makes_ftp_calls(self, filename, text, folder):
        fileset = self._get_ftpfileset()
        csv_stream = StringIO(text)
        mock_ftp = mock.MagicMock()
        with mock.patch.object(fileset, "_get_ftp_connector", return_value=mock_ftp):
            fileset.write_file(filename, csv_stream, folder)
            if folder:
                mock_ftp.cwd.assert_called_once_with(folder)
            mock_ftp.storlines.assert_called_once_with(f"STOR {filename}", mock.ANY)
            mock_ftp.close.assert_called_once()
            csv_bytes = mock_ftp.storlines.call_args.args[1]
            csv_stream_as_bytes = fileset._convert_stringio_to_bytesio(csv_stream)
            assert csv_bytes.getvalue() == csv_stream_as_bytes.getvalue()

    def _get_ftpfileset(self):
        with mock_config(import_csv_folder=None):
            import los_files

            return los_files.FTPFileSet()


class TestSFTPFileSet:
    @given(st.text())
    def test_convert_stringio_to_bytesio(self, text):
        with mock_config(ftp_ssh_key_path="ftp_ssh_key_path"):
            fileset = self._get_sftpfileset()
            stringio = StringIO(text)
            bytesio = fileset._convert_stringio_to_bytesio(stringio)
            assert isinstance(bytesio, BytesIO)

    @settings(deadline=None)
    @given(st.text(), st.text(), st.text())
    def test_makes_sftp_calls(self, filename, text, folder):
        with mock_config(
            ftp_ssh_key_path="ftp_ssh_key_path",
            ftp_ssh_key_pass="ftp_ssh_key_pass",
        ):
            fileset = self._get_sftpfileset()
            csv_stream = StringIO(text)
            mock_sftp = mock.MagicMock()
            mock_ssh = mock.MagicMock()
            with mock.patch.object(
                fileset, "_get_sftp_connector", return_value=(mock_ssh, mock_sftp)
            ):
                fileset.write_file(filename, csv_stream, folder)
                if folder:
                    mock_sftp.chdir.assert_called_once_with(folder)
                mock_sftp.putfo.assert_called_once_with(
                    fl=mock.ANY, remotepath=filename
                )
                mock_sftp.close.assert_called_once()
                csv_bytes = mock_sftp.putfo.call_args.kwargs["fl"]
                csv_stream_as_bytes = fileset._convert_stringio_to_bytesio(csv_stream)
                assert csv_bytes.getvalue() == csv_stream_as_bytes.getvalue()

    def _get_sftpfileset(self):
        with mock_config(import_csv_folder=None, ftp_ssh_key_path="ftp_ssh_key_path"):
            import los_files

            return los_files.SFTPFileSet()



""" This program makes an actual state sqlite-database for use by
jobs run under job-runner as opposed to the actual state used
by customers. It is meant to be run just after the nightly imports
to be used as a speed up in comparison with hitting MO's rest interface.
"""
import logging
import sys

import click
from ra_utils.load_settings import load_settings
from sqlalchemy import create_engine

from .sql_export import SqlExport

LOG_LEVEL = logging.DEBUG

logger = logging.getLogger("lc-for-jobs")


def get_engine(dbpath=None):
    if dbpath is None:
        settings = load_settings()
        dbpath = settings.get("lc-for-jobs.actual_db_name", "ActualState")

    dbpath = str(dbpath)
    if dbpath != ":memory:":
        dbpath += ".db"
    db_string = "sqlite:///{}".format(dbpath)
    return create_engine(db_string)


@click.group()
def cli():
    # Solely used for command grouping
    pass


@cli.command()
@click.option("--resolve-dar/--no-resolve-dar", default=True)
def sql_export(resolve_dar):

    # Load settings file
    settings = load_settings()

    # Override settings
    overrides = {
        "exporters.actual_state.type": "SQLite",
        "exporters.actual_state_historic.type": "SQLite",
        "exporters.actual_state.db_name": settings.get(
            "lc-for-jobs.actual_db_name", "ActualState"
        ),
        "exporters.actual_state.manager_responsibility_class": settings[
            "exporters.actual_state.manager_responsibility_class"
        ],
    }
    settings.update(overrides)

    sql_export = SqlExport(force_sqlite=True, historic=False, settings=settings)
    sql_export.perform_export(resolve_dar=resolve_dar)


if __name__ == "__main__":
    for name in logging.root.manager.loggerDict:  # type: ignore
        if name in ("lc-for-jobs", "LoraCache", "SqlExport"):
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        level=LOG_LEVEL,
        stream=sys.stdout,
    )
    cli()



from mox_helpers.payloads import lora_klasse


def create_org_unit(department, name, org, unit_type, from_date):
    payload = {
        "uuid": department["DepartmentUUIDIdentifier"],
        "user_key": department["DepartmentIdentifier"],
        "name": name,
        "parent": {
            "uuid": department["DepartmentReference"]["DepartmentUUIDIdentifier"]
        },
        "org_unit_type": {"uuid": unit_type},
        "validity": {"from": from_date, "to": None},
    }
    return payload


def create_single_org_unit(department, unit_type, unit_level, parent):
    payload = {
        "uuid": department["DepartmentUUIDIdentifier"],
        "user_key": department["DepartmentIdentifier"],
        "name": department["DepartmentName"],
        "parent": {"uuid": parent},
        "org_unit_type": {"uuid": unit_type},
        "org_unit_level": {"uuid": unit_level},
        "validity": {"from": department["ActivationDate"], "to": None},
    }
    return payload


def edit_org_unit(
    user_key, name, unit_uuid, parent, ou_level, ou_type, from_date, to_date=None
):
    payload = {
        "type": "org_unit",
        "data": {
            "uuid": unit_uuid,
            "user_key": user_key,
            "name": name,
            "parent": {"uuid": parent},
            "org_unit_level": {"uuid": ou_level},
            "org_unit_type": {"uuid": ou_type},
            "validity": {"from": from_date, "to": to_date},
        },
    }
    return payload


def engagement(data, mo_engagement):
    payload = {"type": "engagement", "uuid": mo_engagement["uuid"], "data": data}
    return payload


def association(data, uuid):
    payload = {"type": "association", "uuid": uuid, "data": data}
    return payload


def create_leave(mo_eng, person_uuid, leave_uuid, job_id, validity):
    payload = {
        "type": "leave",
        "engagement": {"uuid": mo_eng["uuid"]},
        "person": {"uuid": person_uuid},
        "leave_type": {"uuid": leave_uuid},
        "user_key": job_id,
        "validity": validity,
    }
    return payload


def create_engagement(
    org_unit,
    person_uuid,
    job_function,
    engagement_type,
    primary,
    user_key,
    engagement_info,
    validity,
    **extensions
):
    try:
        working_time = float(engagement_info["working_time"][0]["OccupationRate"])
    except IndexError:
        working_time = 0
    payload = {
        "type": "engagement",
        "org_unit": {"uuid": org_unit},
        "person": {"uuid": person_uuid},
        "job_function": {"uuid": job_function},
        "primary": {"uuid": primary},
        "engagement_type": {"uuid": engagement_type},
        "user_key": user_key,
        "fraction": int(working_time * 1000000),
        "validity": validity,
    }
    payload.update(extensions)
    return payload


def create_association(org_unit, person_uuid, association_uuid, job_id, validity):
    payload = {
        "type": "association",
        "org_unit": {"uuid": org_unit},
        "person": {"uuid": person_uuid},
        "association_type": {"uuid": association_uuid},
        "user_key": job_id,
        "validity": validity,
    }
    return payload


def connect_it_system_to_user(username, it_system, person_uuid):
    payload = {
        "type": "it",
        "user_key": username,
        "itsystem": {"uuid": it_system},
        "person": {"uuid": person_uuid},
        "validity": {"from": "1930-01-01", "to": None},
    }
    return payload


def edit_klasse_title(titel):
    payload = {
        "attributter": {
            "klasseegenskaber": [
                {
                    "titel": titel,
                    "virkning": {
                        "from": "1930-01-01",
                        "to": "infinity",
                        "aktoerref": "ddc99abd-c1b0-48c2-aef7-74fea841adae",
                        "aktoertypekode": "Bruger",
                    },
                }
            ]
        }
    }
    return payload


def profession(profession, org, job_function_facet):
    klasse = lora_klasse(
        bvn=profession,
        title=profession,
        facet_uuid=job_function_facet,
        org_uuid=org,
        scope="TEXT",
        dato="1930-01-01",
    )
    return klasse


def engagement_type(engagement_type_ref, job_position_id, org, engagement_type_facet):
    klasse = lora_klasse(
        bvn=engagement_type_ref,
        title=job_position_id,
        facet_uuid=engagement_type_facet,
        org_uuid=org,
        scope="TEXT",
        dato="1930-01-01",
    )
    return klasse



import unittest

from parameterized import parameterized

from ..utils import recursive_dict_update


class TestRecursiveDictUpdate(unittest.TestCase):
    @parameterized.expand(
        [
            [
                # Original
                {"alfa": 1},
                # Updates
                {"alfa": 2},
                # Non recursive update
                {"alfa": 2},
                # Recursive update
                {"alfa": 2},
            ],
            [
                # Original
                {"alfa": {"beta": 2, "charlie": 3}},
                # Updates
                {"alfa": {"beta": 4}},
                # Non recursive update
                {"alfa": {"beta": 4}},
                # Recursive update
                {"alfa": {"beta": 4, "charlie": 3}},
            ],
            [
                # Original
                {"alfa": {"beta": {"echo": 4, "fox": 5}, "charlie": 3}},
                # Updates
                {"alfa": {"beta": {"echo": 6}}},
                # Non recursive update
                {"alfa": {"beta": {"echo": 6}}},
                # Recursive update
                {"alfa": {"beta": {"echo": 6, "fox": 5}, "charlie": 3}},
            ],
            [
                # Original
                {"alfa": {"beta": {"echo": 4}, "charlie": {"fox": 5}}},
                # Updates
                {"alfa": {"beta": {"echo": 6}}},
                # Non recursive update
                {"alfa": {"beta": {"echo": 6}}},
                # Recursive update
                {"alfa": {"beta": {"echo": 6}, "charlie": {"fox": 5}}},
            ],
        ]
    )
    def test_function(self, original, updates, expected, expected_recursive):
        updated = {**original, **updates}
        r_updated = recursive_dict_update(original, updates)
        self.assertEqual(updated, expected)
        self.assertEqual(r_updated, expected_recursive)



from sdlon.models import SDBasePerson, EmploymentWithTelephoneNumberIdentifier
from sdlon.sd_to_pydantic import convert_to_sd_base_person


def test_convert_to_sd_base_person__no_contact_info():
    # Arrange
    person = {
        "PersonCivilRegistrationIdentifier": "1111111111",
        "PersonGivenName": "Bruce",
        "PersonSurnameName": "Lee",
        "Employment": {
            "EmploymentIdentifier": "12345",
        },
    }

    # Act
    sd_base_person = convert_to_sd_base_person(person)

    # Assert
    assert sd_base_person == SDBasePerson(
        cpr="1111111111",
        given_name="Bruce",
        surname="Lee",
        emp_with_telephone_number_identifiers=[
            EmploymentWithTelephoneNumberIdentifier(
                employment_identifier="12345", telephone_number_ids=[]
            )
        ],
    )


def test_convert_to_sd_base_person__with_contact_info():
    # Arrange
    person = {
        "PersonCivilRegistrationIdentifier": "1111111111",
        "PersonGivenName": "Bruce",
        "PersonSurnameName": "Lee",
        "Employment": {
            "EmploymentIdentifier": "12345",
            "ContactInformation": {"TelephoneNumberIdentifier": ["12345678", "14"]},
        },
    }

    # Act
    sd_base_person = convert_to_sd_base_person(person)

    # Assert
    assert sd_base_person == SDBasePerson(
        cpr="1111111111",
        given_name="Bruce",
        surname="Lee",
        emp_with_telephone_number_identifiers=[
            EmploymentWithTelephoneNumberIdentifier(
                employment_identifier="12345", telephone_number_ids=["12345678", "14"]
            )
        ],
    )


def test_convert_to_sd_base_person__with_contact_info_multiple_employments():
    # Arrange
    person = {
        "PersonCivilRegistrationIdentifier": "1111111111",
        "PersonGivenName": "Bruce",
        "PersonSurnameName": "Lee",
        "Employment": [
            {
                "EmploymentIdentifier": "12345",
                "ContactInformation": {"TelephoneNumberIdentifier": ["12345678", "14"]},
            },
            {
                "EmploymentIdentifier": "54321",
                "ContactInformation": {"TelephoneNumberIdentifier": ["87654321"]},
            },
        ],
    }

    # Act
    sd_base_person = convert_to_sd_base_person(person)

    # Assert
    assert sd_base_person == SDBasePerson(
        cpr="1111111111",
        given_name="Bruce",
        surname="Lee",
        emp_with_telephone_number_identifiers=[
            EmploymentWithTelephoneNumberIdentifier(
                employment_identifier="12345", telephone_number_ids=["12345678", "14"]
            ),
            EmploymentWithTelephoneNumberIdentifier(
                employment_identifier="54321", telephone_number_ids=["87654321"]
            ),
        ],
    )



import logging
import urllib.parse
from enum import Enum
from functools import partial
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

from ra_utils.load_settings import load_settings


logger = logging.getLogger("SqlExport")


class DatabaseFunction(Enum):
    ACTUAL_STATE = 1
    ACTUAL_STATE_HISTORIC = 2


# Mapping from DatabaseFunction + conceptual key to actual keys in settings.json
SETTINGS_MAP = {
    DatabaseFunction.ACTUAL_STATE: {
        "type": "exporters.actual_state.type",
        "host": "exporters.actual_state.host",
        "db_name": "exporters.actual_state.db_name",
        "username": "exporters.actual_state.user",
        "password": "exporters.actual_state.password",
    },
    DatabaseFunction.ACTUAL_STATE_HISTORIC: {
        "type": "exporters.actual_state_historic.type",
        "host": "exporters.actual_state_historic.host",
        "db_name": "exporters.actual_state_historic.db_name",
        "username": "exporters.actual_state_historic.user",
        "password": "exporters.actual_state_historic.password",
    },
}

# Mapping from DatabaseFunction + conceptual key to fallback keys in settings.json
# These are used if the settings map above does not yield a value
FALLBACK_MAP = {
    DatabaseFunction.ACTUAL_STATE_HISTORIC: {
        "host": "exporters.actual_state.host",
        "username": "exporters.actual_state.user",
        "password": "exporters.actual_state.password",
    },
}


class ConfigurationError(Exception):
    """Throw whenever a configuration issue in settings.json is detected."""

    pass


def convert_key_to_setting(key: str, database_function: DatabaseFunction) -> str:
    """Convert a connection key to an actual setting in settings.json."""
    return SETTINGS_MAP[database_function][key]


def convert_key_to_fallback(
    key: str, database_function: DatabaseFunction
) -> Optional[str]:
    """Convert a connection key to a fallback setting in settings.json."""
    return FALLBACK_MAP.get(database_function, {}).get(key)


def load_setting(
    key: str,
    database_function: DatabaseFunction,
    settings: Dict,
) -> Optional[str]:
    setting = convert_key_to_setting(key, database_function)
    fallback_setting = convert_key_to_fallback(key, database_function)

    value = settings.get(setting)
    if value is None and fallback_setting:
        logger.warning(f"Utilizing {fallback_setting} in place of {setting}!")
        value = settings.get(fallback_setting)
    return value


def get_db_type(
    database_function: DatabaseFunction, force_sqlite: bool, settings: Dict
) -> str:
    if force_sqlite:
        return "SQLite"
    value = load_setting("type", database_function, settings)
    if value is None:
        raise ConfigurationError("Missing value in db_type")
    return value


def get_db_name(database_function: DatabaseFunction, settings: Dict) -> str:
    value = load_setting("db_name", database_function, settings)
    if value is None:
        raise ConfigurationError("Missing value in db_name")
    return value


get_db_host = partial(load_setting, "host")
get_db_username = partial(load_setting, "username")
get_db_password = partial(load_setting, "password")


def generate_db_type_and_name(
    database_function: DatabaseFunction, force_sqlite: bool, settings: Dict
) -> Tuple[str, str]:
    return (
        get_db_type(database_function, force_sqlite, settings),
        get_db_name(database_function, settings),
    )


def generate_connection_url(
    database_function: DatabaseFunction,
    force_sqlite: bool = False,
    settings: Optional[Dict] = None,
) -> str:
    """Utilize settings or settings from disk to derive database connection url."""
    settings = settings or load_settings()

    db_type, db_name = generate_db_type_and_name(
        database_function, force_sqlite, settings
    )
    user = get_db_username(database_function, settings)
    db_host = get_db_host(database_function, settings)
    pw_raw = get_db_password(database_function, settings)
    pw_raw = pw_raw or ""
    pw = urllib.parse.quote_plus(pw_raw)

    if db_type == "Memory":
        return "sqlite://"
    if db_type == "SQLite":
        return "sqlite:///{}.db".format(db_name)
    if db_type == "MS-SQL":
        return "mssql+pymssql://{}:{}@{}/{}".format(user, pw, db_host, db_name)
    if db_type == "MS-SQL-ODBC":
        quoted = urllib.parse.quote_plus(
            (
                "DRIVER=libtdsodbc.so;Server={};Database={};UID={};"
                + "PWD={};TDS_Version=8.0;Port=1433;"
            ).format(db_host, db_name, user, pw_raw)
        )
        return "mssql+pyodbc:///?odbc_connect={}".format(quoted)
    if db_type == "Mysql":
        return "mysql+mysqldb://{}:{}@{}/{}".format(user, pw, db_host, db_name)
    if db_type == "Postgres":
        return "postgresql://{}:{}@{}/{}".format(user, pw, db_host, db_name)
    raise Exception("Unknown DB type")


def generate_engine_settings(
    database_function: DatabaseFunction,
    force_sqlite: bool = False,
    settings: Optional[Dict] = None,
) -> Dict[str, Any]:
    settings = settings or load_settings()

    db_type = get_db_type(database_function, force_sqlite, settings)
    engine_settings: Dict = {"pool_pre_ping": True}
    if db_type == "Mysql":
        engine_settings.update({"pool_recycle": 3600})
    return engine_settings



from datetime import date
from collections import OrderedDict
from unittest.mock import patch

from sdlon.config import CommonSettings
from sdlon.models import JobFunction
from sdlon.sd_common import read_employment_at


@patch("sdlon.sd_common.sd_lookup")
def test_return_none_when_sd_employment_empty(mock_sd_lookup):
    mock_sd_lookup.return_value = OrderedDict()
    settings = CommonSettings(
        sd_global_from_date=date(2000, 1, 1),
        sd_import_run_db="not used",
        sd_institution_identifier="dummy",
        sd_user="user",
        sd_password="password",
        sd_job_function=JobFunction.employment_name,
        sd_monthly_hourly_divide=1,
    )

    assert read_employment_at(date(2000, 1, 1), settings) is None



from functools import lru_cache
from typing import Optional

from os2mo_dar_client import DARClient


@lru_cache(maxsize=None)
def dawa_lookup(street_name: str, postal_code: str) -> Optional[str]:
    """Lookup an address object in DAWA and try to find an UUID for the address.

    Note: `street_name` is a misleading parameter name, as it should contain the street
          name along with house number, door, story, etc.

    Args:
        street_name: Address string without postal code, city or country.
        postal_code: Postal code for the street_name.

    Returns:
        DAWA UUID for the address, or None if it is not uniquely found.
    """
    combined_address_string = f"{street_name}, {postal_code}"

    dar_uuid = None
    try:
        darclient = DARClient()
        with darclient:
            dar_reply = darclient.cleanse_single(combined_address_string)
            dar_uuid = dar_reply["id"]
    except Exception as exp:
        print(exp, " during dawa_lookup")
    return dar_uuid



from sdlon.graphql import get_mo_client


def test_get_mo_client():
    # Act
    client = get_mo_client(
        "http://keycloak-service:8080/auth",
        "client_id",
        "client_secret",
        "http://mo-service:5000",
        10,
        300,
    )

    # Assert
    assert client.transport.url == "http://mo-service:5000/graphql/v10"
    assert (
        client.transport.client_args["auth_server"]
        == "http://keycloak-service:8080/auth"
    )
    assert client.transport.client_args["client_id"] == "client_id"
    assert client.transport.client_args["client_secret"] == "client_secret"
    assert client.transport.client_args["timeout"] == 300



import logging
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List

import config
import uuids
from initial_classes import Class
from initial_classes import CLASSES
from mox_helpers import payloads as mox_payloads
from mox_helpers.mox_helper import create_mox_helper
from mox_helpers.mox_helper import ElementNotFound
from os2mo_data_import import ImportHelper  # type: ignore
from os2mo_data_import.mox_data_types import Itsystem

logger = logging.getLogger(__name__)


@dataclass
class ClassImportResult:
    source: Class
    lora_payload: Dict[str, Any]


async def perform_initial_setup():
    """
    Perform all initial bootstrapping of OS2mo.
    Imports an organisation if missing, and adds all base facets
    Imports all pretedetermined classes and it systems
    """
    settings = config.get_config()
    mox_helper = await create_mox_helper(settings.mox_base)
    try:
        await mox_helper.read_element_organisation_organisation(bvn="%")
    except ElementNotFound:
        logger.info("No org found in LoRa. Performing initial setup.")
        importer = ImportHelper(
            create_defaults=True,
            mox_base=settings.mox_base,
            mora_base=settings.mora_base,
            seperate_names=True,
        )
        importer.add_organisation(
            identifier="Århus Kommune",
            user_key="Århus Kommune",
            municipality_code=751,
            uuid=uuids.ORG_UUID,
        )
        # Perform initial import of org and facets
        importer.import_all()

    await import_remaining_classes()
    await import_it()


async def import_remaining_classes():
    """
    Import a set of predetermined classes. All the classes have predefined UUIDs
    which makes this function idempotent
    """
    settings = config.get_config()
    mox_helper = await create_mox_helper(settings.mox_base)
    result: List[ClassImportResult] = []

    for cls in CLASSES:
        facet_uuid = await mox_helper.read_element_klassifikation_facet(bvn=cls.facet)
        lora_payload = mox_payloads.lora_klasse(
            bvn=cls.bvn,
            title=cls.titel,
            facet_uuid=str(facet_uuid),
            org_uuid=str(uuids.ORG_UUID),
            scope=cls.scope,
        )
        await mox_helper.insert_klassifikation_klasse(lora_payload, str(cls.uuid))
        result.append(ClassImportResult(cls, lora_payload))

    return result


async def import_it():
    """
    Import predetermined IT systems. The UUID(s) are predefined which makes this
    function idempotent.
    """
    settings = config.get_config()
    if settings.azid_it_system_uuid == uuids.AZID_SYSTEM:
        mox_helper = await create_mox_helper(settings.mox_base)
        it_system = Itsystem(system_name="AZ", user_key="AZ")
        it_system.organisation_uuid = str(uuids.ORG_UUID)
        uuid = uuids.AZID_SYSTEM
        json = it_system.build()
        await mox_helper.insert_organisation_itsystem(json, str(uuid))
    else:
        logger.info(
            "Settings specify a non-default AZID IT system UUID, not creating "
            "default AZ IT system"
        )



class ADError(Exception):
    """Superclass for AD related exceptions."""


class EngagementDatesError(ADError):
    """Error to raise when errors in engagement dates occur."""


class NoPrimaryEngagementException(ADError):
    pass


class NoActiveEngagementsException(ADError):
    pass


class UserNotFoundException(ADError):
    pass


class CprNotFoundInADException(ADError):
    pass


class ManagerNotUniqueFromCprException(ADError):
    pass


class CprNotNotUnique(ADError):
    pass


class SamAccountNameNotUnique(ADError):
    pass


class SamAccountDoesNotExist(ADError):
    pass


class ReplicationFailedException(ADError):
    pass


class NoScriptToExecuteException(ADError):
    pass


class UnknownKeywordsInScriptException(ADError):
    pass


class CommandFailure(ADError):
    pass


class ImproperlyConfigured(ADError):
    pass



"""ad_fix_enddate.py

This program takes engagement end dates from MO and writes them to the corresponding AD
user.

The program has evolved from a "one-time fix script" to be responsible for all updates
of end dates in AD. In other words, it has "taken over" writing end dates from
"mo_to_ad_sync.py" (which is unable to "see" past end dates, as it only considers
engagements in the present or future, and thus cannot process terminated engagements.)

The program currently supports two modes of operation:
    a) Writing a single MO end date to a single AD field on each AD user.
    b) Writing two MO end dates to two AD fields on each AD user.

In the first mode, the end date to write is the latest end date of all engagements found
for a given MO user.

In the second mode, the engagement end dates are split into two:
    - the "current" end date is the latest end date occurring in either the past or
      present.
    - the "future" end date is the first end date occurring in the future.
"""
import datetime
import logging
from dataclasses import dataclass
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import Self

import click
import httpx
import sentry_sdk
from dateutil import tz
from fastapi.encoders import jsonable_encoder
from gql import gql
from more_itertools import one
from more_itertools import partition
from ra_utils.job_settings import JobSettings
from ra_utils.load_settings import load_setting
from ra_utils.tqdm_wrapper import tqdm
from raclients.graph.client import GraphQLClient
from raclients.graph.client import SyncClientSession
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_delay
from tenacity import wait_fixed

from integrations.ad_integration.ad_common import AD
from integrations.ad_integration.ad_reader import ADParameterReader


logger = logging.getLogger(__name__)

DEFAULT_TIMEZONE = tz.gettz("Europe/Copenhagen")


def in_default_tz(val: datetime.datetime) -> datetime.datetime:
    return val.astimezone(DEFAULT_TIMEZONE)


class _SymbolicConstant:
    # Helper class which is always equal to itself

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return True
        return super().__eq__(other)


class Unset(_SymbolicConstant):
    """Represents an end date that should *not* be updated in AD.
    Typically because the MO user has no engagements at all, or because the MO user has
    no engagements in the future, or the past/present.

    When encountering an `Unset` end date, the program should not update the matching
    AD field.
    """


class Invalid(_SymbolicConstant):
    """Represents an invalid AD end date, usually due to malformed or unexpected data,
    that cannot be parsed into a `datetime.datetime` object.
    """


class PositiveInfinity(_SymbolicConstant):
    """Represents the upper bound of a validity period if MO returns `None`, i.e. the
    validity period ends at "infinity."
    """

    def as_datetime(self) -> datetime.datetime:
        return in_default_tz(datetime.datetime.fromisoformat("9999-12-31"))


class NegativeInfinity(_SymbolicConstant):
    """Represents the lower bound of a validity period if MO returns `None`, i.e. the
    validity period begins at "-infinity."
    """

    def as_datetime(self) -> datetime.datetime:
        return in_default_tz(datetime.datetime.fromisoformat("1930-01-01"))


ValidityTuple = tuple[
    datetime.datetime | NegativeInfinity,  # from
    datetime.datetime | PositiveInfinity,  # to
]


class AdFixEndDateSettings(JobSettings):
    # Currently only used for calling `.start_logging_based_on_settings` in `cli`

    class Config:
        settings_json_prefix = "integrations.ad.write"


class MOEngagementDateSource:
    """Fetch engagement dates from MO for a given MO user UUID, and return either:
    - a single end date (`get_end_date`), or
    - a current and a future end date (`get_split_end_dates`)
    """

    def __init__(self, graphql_session: SyncClientSession):
        self._graphql_session: SyncClientSession = graphql_session

    @retry(
        wait=wait_fixed(5),
        reraise=True,
        stop=stop_after_delay(10 * 60),
        retry=retry_if_exception_type(httpx.HTTPError),
    )
    def get_employee_engagement_dates(self, uuid: str) -> list[dict]:
        """Fetch all MO engagement validities for a given MO user UUID"""

        query = gql(
            """
            query Get_mo_engagements($employees: [UUID!]) {
                engagements(employees: $employees, from_date: null, to_date: null) {
                    objects {
                        validity {
                            from
                            to
                        }
                    }
                }
            }
            """
        )

        result = self._graphql_session.execute(
            query,
            variable_values=jsonable_encoder({"employees": uuid}),
        )

        if not result["engagements"]:
            raise KeyError("User not found in mo")

        return result["engagements"]

    def get_end_date(self, uuid: str) -> datetime.datetime | PositiveInfinity | Unset:
        """Return a single engagement end date for a given MO user UUID"""

        try:
            engagements = self.get_employee_engagement_dates(uuid)
        except KeyError:
            return Unset()

        parsed_validities = self._parse_validities(engagements)
        if parsed_validities:
            end_date: datetime.datetime | PositiveInfinity = max(
                (validity[1] for validity in parsed_validities),
                key=lambda end_datetime: self._fold_in_utc(
                    end_datetime, datetime.datetime.max
                ),
            )
            return end_date
        else:
            return Unset()

    def get_split_end_dates(
        self, uuid: str
    ) -> tuple[
        datetime.datetime | PositiveInfinity | Unset,
        datetime.datetime | PositiveInfinity | Unset,
    ]:
        """Return a 2-tuple of (current, future) engagement end dates for a given MO
        user UUID.
        """

        def _split(validity):
            """Split the set of engagement validities into two partitions:
            - one partition of past and present engagements.
            - one partition of future engagements.
            The split is determined by looking at the "from" part of the validity.
            """
            local_date = in_default_tz(datetime.datetime.now()).date()
            # Take either engagement start date, or "negative infinity" if blank
            start_date = self._fold_in_utc(validity[0], datetime.datetime.min).date()
            return start_date > local_date

        try:
            engagements = self.get_employee_engagement_dates(uuid)
        except KeyError:
            # If MO user has no engagements, we should not update neither the current
            # nor the future end date.
            return Unset(), Unset()

        parsed_validities: list[ValidityTuple] = self._parse_validities(engagements)

        # Split by whether the engagement *starts* in the future, or not
        current: Iterator[ValidityTuple]
        future: Iterator[ValidityTuple]
        current, future = partition(_split, parsed_validities)

        # Sort by validity end - the validity with the latest end appears first in list
        current_list: list[ValidityTuple] = sorted(
            current,
            key=lambda validity: self._fold_in_utc(validity[1], datetime.datetime.max),
            reverse=True,
        )

        # Sort by validity start - the validity with the earliest start appears first in
        # list.
        future_list: list[ValidityTuple] = sorted(
            future,
            key=lambda validity: self._fold_in_utc(validity[0], datetime.datetime.min),
            reverse=False,
        )

        # Check if either `current` or `future` (or both) is a 0-length list, and return
        # results accordingly.
        if current_list and future_list:
            # Return end date of the latest current validity and the earliest future
            # validity.
            return current_list[0][1], future_list[0][1]
        elif current_list and not future_list:
            # Return only end date of the latest current validity
            return current_list[0][1], Unset()
        elif not current_list and future_list:
            # Return only end date of the earliest future validity
            return Unset(), future_list[0][1]

        return Unset(), Unset()  # no engagements at all

    def _fold_in_utc(
        self,
        val: datetime.datetime | None | PositiveInfinity | NegativeInfinity,
        default: datetime.datetime,
    ) -> datetime.datetime:
        # Fold in UTC as `datetime.datetime.max` and `datetime.datetime.min` cannot be
        # converted into other timezones without surprising issues.
        return (
            default
            if val in (None, PositiveInfinity(), NegativeInfinity())
            else val.astimezone(datetime.timezone.utc).replace(tzinfo=None)  # type: ignore
        )

    def _parse_validities(self, engagements) -> list[ValidityTuple]:
        """Convert validities from string values to 2-tuple of `datetime.datetime`
        objects.
        """

        def _from_iso_or_none(val: str, none: PositiveInfinity | NegativeInfinity):
            if val:
                maybe_naive_dt = datetime.datetime.fromisoformat(val)
                return in_default_tz(maybe_naive_dt)
            # If val is None or '', convert to the `none` value provided by the caller,
            # either `PositiveInfinity` or `NegativeInfinity`.
            return none

        def _start_from_iso_or_none(val: str) -> datetime.datetime | NegativeInfinity:
            return _from_iso_or_none(val, NegativeInfinity())

        def _end_from_iso_or_none(val: str) -> datetime.datetime | PositiveInfinity:
            return _from_iso_or_none(val, PositiveInfinity())

        def _parse(validity: dict) -> ValidityTuple:
            start = _start_from_iso_or_none(validity["from"])
            end = _end_from_iso_or_none(validity["to"])

            # Sanity check that engagement start is indeed before engagement end
            folded_start = self._fold_in_utc(start, datetime.datetime.min)
            folded_end = self._fold_in_utc(end, datetime.datetime.max)
            assert folded_start <= folded_end

            return start, end

        parsed_validities = [
            _parse(obj["validity"])
            for engagement in engagements
            for obj in engagement["objects"]
        ]

        return parsed_validities


@dataclass
class ADUserEndDate:
    """Represents an existing AD end date."""

    mo_uuid: str
    field_name: str
    field_value: str | None | Invalid

    @property
    def normalized_value(self) -> datetime.datetime | None | Invalid:
        if self.field_value == Invalid():
            return Invalid()
        try:
            return in_default_tz(datetime.datetime.fromisoformat(self.field_value))  # type: ignore
        except (TypeError, ValueError):
            logger.debug("cannot parse %r as ISO datetime", self.field_value)
            return Invalid()


class ADEndDateSource:
    """Fetch AD end date(s) for one or all AD users.

    In case we are updating more than one AD end date (`enddate_field_future` is set),
    this class will yield two `ADUserEndDate` instances per AD user.

    In case we are only updating one AD end date  (`enddate_field_future` is None),
    this class will yield one `ADUserEndDate` instances per AD user.
    """

    def __init__(
        self,
        uuid_field: str,
        enddate_field: str,
        enddate_field_future: str | None,
        settings: dict | None = None,
    ):
        self._uuid_field = uuid_field
        self._enddate_field = enddate_field
        self._enddate_field_future = enddate_field_future
        self._reader = ADParameterReader(all_settings=settings)

        # This holds the "raw" result of calling either `ADParameterReader.read_it_all`
        # (in `of_all_users`), or `ADParameterReader.read_user` (in `of_one_user`.)
        self._ad_users: list[dict] = []

    def __iter__(self) -> Iterator[ADUserEndDate]:
        """Yield one or two `ADUserEndDate` instances for each AD user in
        `self._ad_users`.

        Iterating over an `ADEndDateSource` only makes sense after calling either
        `of_all_users` or `of_one_user`.
        """
        for ad_user in self._ad_users:
            if self._uuid_field not in ad_user:
                click.echo(
                    f"User with {ad_user['ObjectGuid']=} does not have an "
                    f"{self._uuid_field} field, and will be skipped"
                )
                continue

            yield ADUserEndDate(
                ad_user[self._uuid_field],
                self._enddate_field,
                self._get_case_insensitive(ad_user, self._enddate_field),
            )

            if self._enddate_field_future:
                yield ADUserEndDate(
                    ad_user[self._uuid_field],
                    self._enddate_field_future,
                    self._get_case_insensitive(ad_user, self._enddate_field_future),
                )

    def of_all_users(self) -> Self:
        """Return `ADEndDateSource` for all AD users"""
        self._ad_users = self._reader.read_it_all()
        return self

    def of_one_user(self, username: str) -> Self:
        """Return `ADEndDateSource` for a single AD user given by its AD `username`"""
        self._ad_users = [self._reader.read_user(user=username)]
        return self

    def _get_case_insensitive(self, ad_user: dict, field_name: str) -> str | Invalid:
        # There may be case inconsistencies in AD field names between the name we ask
        # for, and the name that is actually returned. E.g. asking for a field name in
        # all lowercase may return the same field but with a mixed-case name.
        # This method handles this by treating both field names as lowercase.
        #
        # If the field name is not present in the given AD user, an `Invalid` instance
        # is returned.
        try:
            return one(
                (v for k, v in ad_user.items() if k.lower() == field_name.lower())
            )
        except ValueError:
            return Invalid()


class CompareEndDate:
    """Compares the end dates in `mo_engagement_date_source` and
    `ad_end_date_source`, and produces a list of changes that need to be made in AD
    (`get_changes`.)
    """

    def __init__(
        self,
        enddate_field: str,
        enddate_field_future: str | None,
        mo_engagement_date_source: MOEngagementDateSource,
        ad_end_date_source: ADEndDateSource,
    ):
        self._enddate_field = enddate_field
        self._enddate_field_future = enddate_field_future
        self._mo_engagement_date_source = mo_engagement_date_source
        self._ad_end_date_source = ad_end_date_source

    def get_results(
        self,
    ) -> Iterator[tuple[ADUserEndDate, datetime.datetime | Unset | PositiveInfinity]]:
        """For each AD user end date in `self._ad_end_date_source`, produce the relevant
        MO end date(s).
        """
        for ad_user in self._ad_end_date_source:
            if self._enddate_field_future:
                # Mode two: we are updating a "current" and "future" end date in AD
                split = self._mo_engagement_date_source.get_split_end_dates(
                    ad_user.mo_uuid
                )
                current_mo_end_date, future_mo_end_date = split
                if ad_user.field_name == self._enddate_field:
                    yield ad_user, current_mo_end_date
                if ad_user.field_name == self._enddate_field_future:
                    yield ad_user, future_mo_end_date
            else:
                # Mode one: we are updating a single end date in AD
                end_date = self._mo_engagement_date_source.get_end_date(ad_user.mo_uuid)
                if ad_user.field_name == self._enddate_field:
                    yield ad_user, end_date

    def get_changes(self) -> Iterator[tuple[ADUserEndDate, datetime.datetime]]:
        """For each pair of (AD user end date, MO end date), compare the two, and
        produce a "hit" if the AD user end date differs from the MO end date, and thus
        needs to be updated.
        """
        for ad_user, mo_value in self.get_results():
            ad_value = ad_user.normalized_value

            # Convert `None` to "max date"
            if mo_value is None:
                mo_value = PositiveInfinity().as_datetime()

            # Convert `PositiveInfinity` and `NegativeInfinity` to their respective
            # `datetime.datetime` values.
            if mo_value in (PositiveInfinity(), NegativeInfinity()):
                mo_value = mo_value.as_datetime()  # type: ignore

            if mo_value == Unset():
                # If MO value is `Unset`, leave it out of the set of changes
                logger.info(
                    "MO user %r: skipping %r update as MO value is unset",
                    ad_user.mo_uuid,
                    ad_user.field_name,
                )
            elif ad_value != mo_value:
                # If MO and AD values differ, include in the list of changes
                yield ad_user, mo_value  # type: ignore
            else:
                # MO and AD values must be equal
                logger.debug(
                    "MO user %r: normalized MO and AD values are identical in field %r "
                    "(values in MO: %r, AD: %r)",
                    ad_user.mo_uuid,
                    ad_user.field_name,
                    mo_value.strftime("%Y-%m-%d"),  # type: ignore
                    ad_user.field_value,
                )


class UpdateEndDate(AD):
    """Given a list of changes, perform one or two update commands against AD, to bring
    the end dates of the AD users up-to-date.
    """

    def __init__(self, settings=None):
        super().__init__(all_settings=settings)

    def get_update_cmd(
        self,
        uuid_field: str,
        uuid: str,
        end_date_field: str,
        end_date: str,
    ) -> str:
        """Return the relevant Powershell update command to update the given end date
        field in AD with the given value."""

        cmd_f = """
        Get-ADUser %(complete)s -Filter '%(uuid_field)s -eq "%(uuid)s"' |
        Set-ADUser %(credentials)s -Replace @{%(enddate_field)s="%(end_date)s"} |
        ConvertTo-Json
        """
        cmd = cmd_f % dict(
            uuid=uuid,
            end_date=end_date,
            enddate_field=end_date_field,
            uuid_field=uuid_field,
            complete=self._ps_boiler_plate()["complete"],
            credentials=self._ps_boiler_plate()["credentials"],
        )
        return cmd

    def run(self, cmd) -> dict:
        """Run a PowerShell command against AD"""
        return self._run_ps_script("%s\n%s" % (self._build_user_credential(), cmd))

    def run_all(
        self,
        changes: Iterable[tuple[ADUserEndDate, datetime.datetime]],
        uuid_field: str,
        dry: bool = False,
    ) -> list[tuple[str, Any]]:
        """Take the output of `CompareEndDate.get_changes` and turn it into update
        commands. Run the update commands against AD if `dry` is False, or else just log
        the commands that would have been run.
        """

        changes = tqdm(list(changes))
        num_changes = 0
        retval = []

        for ad_user, mo_value in changes:
            cmd = self.get_update_cmd(
                uuid_field,
                ad_user.mo_uuid,
                ad_user.field_name,
                mo_value.strftime("%Y-%m-%d"),
            )
            logger.info(
                "Updating AD user %r, %r = %r",
                ad_user.mo_uuid,
                ad_user.field_name,
                mo_value.strftime("%Y-%m-%d"),
            )
            logger.debug(cmd)
            if dry:
                retval.append((cmd, "<dry run>"))
            else:
                result = self.run(cmd)
                retval.append((cmd, result))  # type: ignore
                if result != {}:
                    logger.error("AD error response %r", result)
                else:
                    num_changes += 1

        logger.info("%d users end dates corrected", num_changes)
        logger.info("All end dates are fixed")

        return retval


@click.command()
@click.option(
    "--enddate-field",
    default=load_setting("integrations.ad_writer.fixup_enddate_field"),
)
@click.option(
    "--enddate-field-future",
    default=load_setting("integrations.ad_writer.fixup_enddate_field_future", None),
)
@click.option("--uuid-field", default=load_setting("integrations.ad.write.uuid_field"))
@click.option(
    "--ad-user",
    help="If given, update only one AD user (specified by username)",
)
@click.option("--dry-run", is_flag=True)
@click.option("--mora-base", envvar="MORA_BASE", default="http://mo")
@click.option("--client-id", envvar="CLIENT_ID", default="dipex")
@click.option("--client-secret", envvar="CLIENT_SECRET")
@click.option("--auth-realm", envvar="AUTH_REALM", default="mo")
@click.option("--auth-server", envvar="AUTH_SERVER", default="http://keycloak")
def cli(
    enddate_field,
    enddate_field_future,
    uuid_field,
    ad_user,
    dry_run,
    mora_base: str,
    client_id: str,
    client_secret: str,
    auth_realm: str,
    auth_server: str,
):
    """Writes MO engagements end date(s) to one or more fields on the corresponding AD
    user(s).
    """
    pydantic_settings = AdFixEndDateSettings()
    pydantic_settings.start_logging_based_on_settings()

    if pydantic_settings.sentry_dsn:
        sentry_sdk.init(dsn=pydantic_settings.sentry_dsn)

    logger.info(
        f"Command line args:"
        f" end-date-field = {enddate_field},"
        f" end-date-field-future = {enddate_field_future},"
        f" uuid-field = {uuid_field},"
        f" dry-run = {dry_run},"
        f" mora-base = {mora_base},"
        f" client-id = {client_id},"
        f" client-secret = not logged,"
        f" auth-realm = {auth_realm},"
        f" auth-server = {auth_server}",
    )

    graphql_client = GraphQLClient(
        url=f"{mora_base}/graphql/v3",
        client_id=client_id,
        client_secret=client_secret,
        auth_realm=auth_realm,
        auth_server=auth_server,
        sync=True,
        httpx_client_kwargs={"timeout": None},
    )

    ad_end_date_source = ADEndDateSource(
        uuid_field,
        enddate_field,
        enddate_field_future,
    )
    if ad_user:
        ad_end_date_source = ad_end_date_source.of_one_user(ad_user)
    else:
        ad_end_date_source = ad_end_date_source.of_all_users()

    with graphql_client as session:
        mo_engagement_date_source = MOEngagementDateSource(session)
        compare = CompareEndDate(
            enddate_field,
            enddate_field_future,
            mo_engagement_date_source,
            ad_end_date_source,
        )
        update = UpdateEndDate()
        update.run_all(compare.get_changes(), uuid_field, dry=dry_run)


if __name__ == "__main__":
    cli()



# At the time of writing, the SD importer creates engagements in MO
# even though the corresponding EmploymentStatusCode is 8 ("ophørt")
# in the XML payload from SD. This script fixes this, i.e. the script
# will:
#
# 1) Get all status 8 employments from SD
# 2) Iterate over these and terminate the corresponding active engagements in MO

import pathlib
import pickle
from datetime import datetime
from typing import List
from uuid import UUID

import click
from gql import gql
from more_itertools import one, exactly_n
from raclients.graph.client import GraphQLClient

from sdlon.graphql import get_mo_client
from sdclient.client import SDClient
from sdclient.requests import GetEmploymentRequest
from sdclient.responses import GetEmploymentResponse
from sdclient.responses import Person
from ramodels.mo.employee import Employee

def get_sd_employments(
    username: str, password: str, institution_identifier: str
) -> GetEmploymentResponse:
    """
    Get all passive employments from SD (the query params are very
    specific for what is needed here).

    Args:
        username: the username for the SD API
        password: the password for the SD API
        institution_identifier: the SD institution identifier

    Returns:
        The SD employments
    """

    sd_client = SDClient(username, password)
    sd_employments = sd_client.get_employment(
        GetEmploymentRequest(
            InstitutionIdentifier=institution_identifier,
            EffectiveDate=datetime.now().date(),
            StatusActiveIndicator=False,
            StatusPassiveIndicator=True,
            EmploymentStatusIndicator=True,
        )
    )
    return sd_employments


def get_mo_employees(gql_client: GraphQLClient) -> List[Employee]:
    """
    Get all MO employees

    Args:
        gql_client: the GraphQL client

    Returns:
        List of CPR numbers
    """

    query = gql("""
    query GetEmployees {
        employees {
            objects {
                cpr_no
            }
            uuid
        }
    }
    """)
    r = gql_client.execute(query)

    employees = []
    for employee in r["employees"]:
        try:
            employees.append(
                Employee(
                    cpr_no=one(employee["objects"])["cpr_no"],
                    uuid=employee["uuid"]
                )
            )
        except ValueError:
            print("Found invalid CPR!")
            print(employee)

    return employees


def terminate_engagement(
        gql_client: GraphQLClient,
        engagement_uuid: str,
        termination_date: str
) -> None:
    """
    Terminate a MO engagement.

    Args:
        gql_client: the GraphQL client
        engagement_uuid: UUID of the engagement to terminate
        termination_date: the last day of work for the engagement
    """
    graphql_terminate_engagement = gql(
        """
            mutation TerminateEngagement($input: EngagementTerminateInput!) {
                engagement_terminate(input: $input) {
                    uuid
                }
            }
        """
    )

    gql_client.execute(graphql_terminate_engagement, variable_values={
        "input": {
            "uuid": str(engagement_uuid),
            "to": termination_date
        }
    })


def get_mo_engagements(
    gql_client: GraphQLClient, employee_uuid: UUID
) -> list[dict[str, str]]:
    """
    Get MO engagements for a given employee.

    Args:
        gql_client: the GraphQL client
        employee_uuid: UUID of the employee to get the engagements from

    Returns:
        List of dicts where each dict contains user keys, "from date" and
        UUID of the engagements
    """

    query = gql("""
        query GetEngagements($uuid: [UUID!]!) {
            engagements(employees: $uuid) {
                objects {
                    user_key
                    validity {
                        from
                    }
                }
                uuid
            }
        }
    """)

    r = gql_client.execute(query, variable_values={"uuid": str(employee_uuid)})
    engagements = [
        {
            "uuid": engagement["uuid"],
            "user_key": one(engagement["objects"])["user_key"],
            # Convert back and forth between datetime objects and strings?
            # Nah - it is much easier to just use [:10] for this use case
            "from": one(engagement["objects"])["validity"]["from"][:10]
        }
        for engagement in r["engagements"]
    ]
    return engagements


def has_sd_status8(
    sd_employments: GetEmploymentResponse, cpr: str, employment_identifier: str
) -> bool:
    """
    Return True if the MO employee with the given cpr number and
    employment_identifier is in sd_employments

    Args:
        sd_employments: the passive SD employments
        cpr: the CPR number of the employee
        employment_identifier: the SD employment identifier

    Returns:
        True if the combination of CPR and employment identifier has status 8
        in SD and False otherwise
    """
    def has_cpr_and_employment_identifier(person: Person) -> bool:
        cpr_match = cpr == person.PersonCivilRegistrationIdentifier
        employment_identifier_match = exactly_n(person.Employment, 1, lambda emp: emp.EmploymentIdentifier == employment_identifier)
        return cpr_match and employment_identifier_match

    return exactly_n(sd_employments.Person, 1, has_cpr_and_employment_identifier)


@click.command()
@click.option(
    "--username",
    "username",
    type=click.STRING,
    envvar="SD_USER",
    required=True,
    help="SD username"
)
@click.option(
    "--password",
    "password",
    type=click.STRING,
    envvar="SD_PASSWORD",
    required=True,
    help="SD password"
)
@click.option(
    "--institution-identifier",
    "institution_identifier",
    type=click.STRING,
    envvar="SD_INSTITUTION_IDENTIFIER",
    required=True,
    help="SD institution identifier"
)
@click.option(
    "--auth-server",
    "auth_server",
    type=click.STRING,
    default="http://localhost:8090/auth",
    help="Keycloak auth server URL"
)
@click.option(
    "--client-id",
    "client_id",
    type=click.STRING,
    default="dipex",
    help="Keycloak client id"
)
@click.option(
    "--client-secret",
    "client_secret",
    type=click.STRING,
    required=True,
    help="Keycloak client secret for the DIPEX client"
)
@click.option(
    "--mo-base-url",
    "mo_base_url",
    type=click.STRING,
    default="http://localhost:5000",
    help="Base URL for calling MO"
)
@click.option(
    "--use-pickle",
    "use_pickle",
    is_flag=True,
    help="Store SD response locally with pickle and use pickled response "
         "in later runs (useful to avoid unnecessary load on SD during "
         "development)"
)
@click.option(
    "--dry-run",
    "dry_run",
    is_flag=True,
    help="Do not perform any changes in MO"
)
def main(
    username: str,
    password: str,
    institution_identifier: str,
    auth_server: str,
    client_id: str,
    client_secret: str,
    mo_base_url: str,
    use_pickle: bool,
    dry_run: bool,
):
    # Get the SD status employments
    if use_pickle:
        pickle_file = "/tmp/sd_employments.bin"
        if not pathlib.Path(pickle_file).is_file():
            sd_employments = get_sd_employments(
                username, password, institution_identifier
            )
            with open(pickle_file, "bw") as fp:
                pickle.dump(sd_employments, fp)
        with open(pickle_file, "br") as fp:
            sd_employments = pickle.load(fp)
    else:
        sd_employments = get_sd_employments(
            username, password, institution_identifier
        )

    print("Number of SD employments:", len(sd_employments.Person))

    gql_client = get_mo_client(
        auth_server, client_id, client_secret, mo_base_url, 3
    )
    employees = get_mo_employees(gql_client)

    print("Terminate engagements")
    for employee in employees:
        engagements = get_mo_engagements(gql_client, employee.uuid)
        for eng in engagements:
            terminate = has_sd_status8(sd_employments, employee.cpr_no, eng["user_key"])
            if terminate:
                print(employee.cpr_no, eng["user_key"], terminate)
                if not dry_run:
                    terminate_engagement(gql_client, eng["uuid"], eng["from"])


if __name__ == "__main__":
    main()



import datetime
import logging
from typing import Iterator
from unittest.mock import Mock
from unittest.mock import patch

import pytest as pytest
from click.testing import CliRunner
from hypothesis import given
from hypothesis import strategies as st

from ..ad_fix_enddate import ADEndDateSource
from ..ad_fix_enddate import ADUserEndDate
from ..ad_fix_enddate import cli
from ..ad_fix_enddate import CompareEndDate
from ..ad_fix_enddate import DEFAULT_TIMEZONE
from ..ad_fix_enddate import Invalid
from ..ad_fix_enddate import MOEngagementDateSource
from ..ad_fix_enddate import NegativeInfinity
from ..ad_fix_enddate import PositiveInfinity
from ..ad_fix_enddate import Unset
from ..ad_fix_enddate import UpdateEndDate
from ..ad_reader import ADParameterReader
from .mocks import AD_UUID_FIELD
from .mocks import MO_UUID
from .mocks import MockADParameterReader
from .mocks import MockADParameterReaderWithMOUUID


ENDDATE_FIELD = "enddate_field"
ENDDATE_FIELD_FUTURE = "enddate_field_future"
TEST_SEARCH_BASE = "search_base"
TEST_SETTINGS = {
    "primary": {
        "search_base": TEST_SEARCH_BASE,
        "system_user": "username",
        "password": "password",
    },
}
VALID_AD_DATE = "2020-01-01"
VALID_AD_DATE_FUTURE = "2025-01-01"
INVALID_AD_DATE = "<invalid date>"


class MockADParameterReaderWithMOUUIDAndInvalidDate(MockADParameterReaderWithMOUUID):
    def read_user(self, **kwargs):
        ad_user = super().read_user(**kwargs)
        ad_user[ENDDATE_FIELD] = INVALID_AD_DATE
        return ad_user


class MockADParameterReaderWithMOUUIDAndValidSplitDates(
    MockADParameterReaderWithMOUUID
):
    def read_user(self, **kwargs):
        ad_user = super().read_user(**kwargs)
        ad_user[ENDDATE_FIELD] = VALID_AD_DATE
        ad_user[ENDDATE_FIELD_FUTURE] = VALID_AD_DATE_FUTURE
        return ad_user


class _MockADEndDateSource(ADEndDateSource):
    def __init__(self):
        # Explicitly avoid calling `super().__init__(...)` as we don't want to create an
        # `ADParameterReader` instance.
        pass

    def __iter__(self) -> Iterator[ADUserEndDate]:
        raise NotImplementedError("must be implemented by subclass")


class _MockADEndDateSourceNoMatchingADUser(_MockADEndDateSource):
    def __iter__(self) -> Iterator[ADUserEndDate]:
        return iter([])


class _MockADEndDateSourceMatchingADUser(_MockADEndDateSource):
    def __iter__(self) -> Iterator[ADUserEndDate]:
        yield ADUserEndDate(MO_UUID, ENDDATE_FIELD, None)


class _MockADEndDateSourceMatchingADUserAndEndDate(_MockADEndDateSource):
    def __iter__(self) -> Iterator[ADUserEndDate]:
        yield ADUserEndDate(MO_UUID, ENDDATE_FIELD, VALID_AD_DATE)


class _MockADEndDateSourceMatchingADUserWrongEndDate(_MockADEndDateSource):
    def __iter__(self) -> Iterator[ADUserEndDate]:
        yield ADUserEndDate(MO_UUID, ENDDATE_FIELD, "2023-01-01")


class _MockADEndDateSourceMatchingADUserWithSplitEndDates(_MockADEndDateSource):
    def __iter__(self) -> Iterator[ADUserEndDate]:
        yield ADUserEndDate(MO_UUID, ENDDATE_FIELD, VALID_AD_DATE)
        yield ADUserEndDate(MO_UUID, ENDDATE_FIELD_FUTURE, VALID_AD_DATE_FUTURE)


class _TestableCompareEndDate(CompareEndDate):
    def __init__(
        self,
        mo_engagement_date_source: MOEngagementDateSource,
        ad_end_date_source: ADEndDateSource,
        enddate_field_future: str | None = None,
    ):
        super().__init__(
            ENDDATE_FIELD,
            enddate_field_future,
            mo_engagement_date_source,
            ad_end_date_source,
        )


class _TestableUpdateEndDate(UpdateEndDate):
    def __init__(self):
        super().__init__(settings=TEST_SETTINGS)
        self._ps_scripts_run = []

    def _run_ps_script(self, ps_script):
        self._ps_scripts_run.append(ps_script)
        return {}


class _TestableUpdateEndDateReturningError(_TestableUpdateEndDate):
    def _run_ps_script(self, ps_script):
        super()._run_ps_script(ps_script)
        return {"error": "is mocked"}  # non-empty dict indicates a Powershell error


def dt(val: str):
    return datetime.datetime.fromisoformat(val).astimezone(DEFAULT_TIMEZONE)


def validity(start: str, end: str | None) -> dict:
    from_date = dt(start)
    to_date = dt(end) if end else None
    return {
        "validity": {
            "from": str(from_date),
            "to": str(to_date) if to_date else None,
        }
    }


def engagement_objects(*validities: dict) -> dict:
    return {"engagements": [{"objects": validities}]}


def _get_mock_graphql_session(return_value):
    graphql_session = Mock()
    graphql_session.execute = Mock()
    graphql_session.execute.return_value = return_value
    return graphql_session


@pytest.fixture()
def mock_graphql_session():
    return _get_mock_graphql_session(
        {
            "engagements": [
                {"objects": [{"validity": {"from": "2020-01-01", "to": VALID_AD_DATE}}]}
            ]
        }
    )


@pytest.fixture()
def mock_graphql_session_raising_keyerror():
    return _get_mock_graphql_session({})


@pytest.fixture()
def mock_mo_engagement_date_source(mock_graphql_session):
    return MOEngagementDateSource(mock_graphql_session)


@pytest.fixture()
def mock_mo_engagement_date_source_raising_keyerror(
    mock_graphql_session_raising_keyerror,
):
    return MOEngagementDateSource(mock_graphql_session_raising_keyerror)


@given(st.builds(Unset), st.builds(Unset))
def test_unset_class_always_equals_itself(unset_a: Unset, unset_b: Unset):
    assert unset_a == unset_b


@given(st.builds(Unset), st.datetimes() | st.none())
def test_unset_class_never_equals_other_types(unset: Unset, other: dict | None):
    assert unset != other


@given(st.builds(Unset))
def test_unset_repr_is_constant(unset: Unset):
    assert repr(unset) == "Unset()"


def test_mo_engagement_date_source_raises_keyerror_on_no_engagements():
    mock_mo_engagement_date_source = MOEngagementDateSource(
        _get_mock_graphql_session({"engagements": []})
    )
    with pytest.raises(KeyError):
        mock_mo_engagement_date_source.get_employee_engagement_dates(MO_UUID)


@pytest.mark.parametrize(
    "eng",
    [
        {
            "engagements": [
                {
                    "objects": [
                        {
                            "validity": {
                                "from": "2020-09-02T00:00:00+02:00",
                                "to": "2021-09-02T00:00:00+02:00",
                            }
                        }
                    ]
                },
                {
                    "objects": [
                        {
                            "validity": {
                                "from": "2021-09-02T00:00:00+02:00",
                                "to": "2022-09-02T00:00:00+02:00",
                            }
                        }
                    ]
                },
                {
                    "objects": [
                        {
                            "validity": {
                                "from": "2022-09-02T00:00:00+02:00",
                                "to": "2023-09-02T00:00:00+02:00",
                            }
                        }
                    ]
                },
            ]
        },
        {
            "engagements": [
                {
                    "objects": [
                        {
                            "validity": {
                                "from": "2020-09-02T00:00:00+02:00",
                                "to": "2021-09-02T00:00:00+02:00",
                            }
                        },
                        {
                            "validity": {
                                "from": "2021-09-02T00:00:00+02:00",
                                "to": "2022-09-02T00:00:00+02:00",
                            }
                        },
                        {
                            "validity": {
                                "from": "2022-09-02T00:00:00+02:00",
                                "to": "2023-09-02T00:00:00+02:00",
                            }
                        },
                    ]
                }
            ]
        },
    ],
)
def test_get_end_date(eng):
    mo_engagement_date_source = MOEngagementDateSource(_get_mock_graphql_session(eng))
    expected_latest_date = datetime.date(2023, 9, 2)
    actual_latest_date = mo_engagement_date_source.get_end_date(MO_UUID).date()
    assert actual_latest_date == expected_latest_date


def test_get_end_date_returns_unset_on_no_engagements():
    mo_engagement_date_source = MOEngagementDateSource(
        _get_mock_graphql_session(engagement_objects())
    )
    assert mo_engagement_date_source.get_end_date(MO_UUID) == Unset()


@pytest.mark.parametrize(
    "mock_response,expected_split",
    [
        # Case A1: only past and present engagements; latest engagement has a blank end
        # date.
        (
            engagement_objects(
                validity("2020-01-01", "2021-12-31"),
                validity("2022-01-01", None),
            ),
            (
                PositiveInfinity(),
                Unset(),
            ),
        ),
        # Case A2: only past and present engagements; latest engagement has a non-blank
        # end date.
        (
            engagement_objects(
                validity("2020-01-01", "2021-12-31"),
                validity("2022-01-01", "2022-12-31"),
            ),
            (
                dt("2022-12-31"),
                Unset(),
            ),
        ),
        # Case B1: only future engagements; earliest engagement has a blank end date
        (
            engagement_objects(
                validity("2024-01-01", None),
                validity("2025-01-01", None),
            ),
            (
                Unset(),
                PositiveInfinity(),
            ),
        ),
        # Case B2: only future engagements; earliest engagement has an end date
        (
            engagement_objects(
                validity("2024-01-01", "2024-12-31"),
                validity("2025-01-01", None),
            ),
            (
                Unset(),
                dt("2024-12-31"),
            ),
        ),
        # Case C1: engagements in the past, present and future. Both the latest
        # engagement in the present, and the earliest engagement in the future, have a
        # blank end date.
        (
            engagement_objects(
                validity("2020-01-01", "2021-12-31"),
                validity("2022-01-01", None),
                validity("2024-01-01", None),
                validity("2025-01-01", None),
            ),
            (
                PositiveInfinity(),
                PositiveInfinity(),
            ),
        ),
        # Case C2: engagements in the past, present and future. The latest engagement in
        # the present has a blank end date, while the earliest engagement in the future
        # has a non-blank end date.
        (
            engagement_objects(
                validity("2020-01-01", "2021-12-31"),
                validity("2022-01-01", None),
                validity("2024-01-01", "2024-12-31"),
                validity("2025-01-01", None),
            ),
            (
                PositiveInfinity(),
                dt("2024-12-31"),
            ),
        ),
        # Case C3: engagements in the past, present and future. The latest engagement in
        # the present has a non-blank end date, while the earliest engagement in the
        # future has a blank end date.
        (
            engagement_objects(
                validity("2020-01-01", "2021-12-31"),
                validity("2022-01-01", "2022-12-31"),
                validity("2024-01-01", None),
                validity("2025-01-01", None),
            ),
            (
                dt("2022-12-31"),
                PositiveInfinity(),
            ),
        ),
    ],
)
def test_get_split_end_dates(mock_response, expected_split):
    mo_engagement_date_source = MOEngagementDateSource(
        _get_mock_graphql_session(mock_response)
    )
    actual_split = mo_engagement_date_source.get_split_end_dates(MO_UUID)
    assert actual_split == expected_split


@pytest.mark.parametrize(
    "engagement_objects",
    [
        {"engagements": []},  # no engagements
        engagement_objects(),  # `engagements` key is present but `objects` is empty
    ],
)
def test_get_split_end_dates_returns_unset_tuple_on_no_engagements(engagement_objects):
    mock_mo_engagement_date_source = MOEngagementDateSource(
        _get_mock_graphql_session(engagement_objects)
    )
    expected_result = Unset(), Unset()
    actual_result = mock_mo_engagement_date_source.get_split_end_dates(MO_UUID)
    assert actual_result == expected_result


@pytest.mark.parametrize(
    "mock_engagements,mock_ad_enddate_source,expected_result",
    [
        # Case 1a: we are updating only the normal end date in AD, and the user only has
        # one engagement (in the past.) We should write the end date of that engagement
        # to the regular end date field in AD.
        (
            engagement_objects(validity(VALID_AD_DATE, VALID_AD_DATE)),
            _MockADEndDateSourceMatchingADUserAndEndDate(),
            [
                (
                    ADUserEndDate(MO_UUID, ENDDATE_FIELD, VALID_AD_DATE),
                    dt(VALID_AD_DATE),
                ),
            ],
        ),
        # Case 1b: we are updating only the normal end date in AD, and the user only has
        # one engagement (in the past), whose end date is blank. We should write the end
        # date of that engagement to the regular end date field in AD.
        (
            engagement_objects(validity(VALID_AD_DATE, None)),
            _MockADEndDateSourceMatchingADUserAndEndDate(),
            [
                (
                    ADUserEndDate(MO_UUID, ENDDATE_FIELD, VALID_AD_DATE),
                    PositiveInfinity(),
                ),
            ],
        ),
        # Case 2a: we are updating both the normal and the future end date in AD, and
        # the user only has one engagement (in the past.) We should write the end date
        # of that engagement to the regular end date field in AD, and leave the future
        # end date field in AD unset.
        (
            engagement_objects(validity(VALID_AD_DATE, VALID_AD_DATE)),
            _MockADEndDateSourceMatchingADUserWithSplitEndDates(),
            [
                (
                    ADUserEndDate(MO_UUID, ENDDATE_FIELD, VALID_AD_DATE),
                    dt(VALID_AD_DATE),
                ),
                (
                    ADUserEndDate(MO_UUID, ENDDATE_FIELD_FUTURE, VALID_AD_DATE_FUTURE),
                    Unset(),
                ),
            ],
        ),
        # Case 2b: we are updating both the normal and the future end date in AD, and
        # the user only has one engagement (in the past), whose end date is blank.  We
        # should write the end date of that engagement to the regular end date field in
        # AD, and leave the future end date field in AD unset.
        (
            engagement_objects(validity(VALID_AD_DATE, None)),
            _MockADEndDateSourceMatchingADUserWithSplitEndDates(),
            [
                (
                    ADUserEndDate(MO_UUID, ENDDATE_FIELD, VALID_AD_DATE),
                    PositiveInfinity(),
                ),
                (
                    ADUserEndDate(MO_UUID, ENDDATE_FIELD_FUTURE, VALID_AD_DATE_FUTURE),
                    Unset(),
                ),
            ],
        ),
    ],
)
def test_get_results(mock_engagements, mock_ad_enddate_source, expected_result):
    mock_mo_engagement_date_source = MOEngagementDateSource(
        _get_mock_graphql_session(mock_engagements)
    )
    instance = _TestableCompareEndDate(
        mock_mo_engagement_date_source,
        mock_ad_enddate_source,
        enddate_field_future=ENDDATE_FIELD_FUTURE,
    )
    actual_result = list(instance.get_results())
    assert actual_result == expected_result


@pytest.mark.parametrize(
    "input_mo_value,expected_mo_value",
    [
        # MO and AD values are identical, yield nothing
        (dt(VALID_AD_DATE), []),
        # MO and AD values differ, yield MO value
        (dt("2023-01-01"), dt("2023-01-01")),
        # MO value is None, yield "max date"
        (None, PositiveInfinity().as_datetime()),
        # MO value is positive infinity, yield its datetime value
        (PositiveInfinity(), PositiveInfinity().as_datetime()),
        # MO value is negative infinity, yield its datetime value
        (NegativeInfinity(), NegativeInfinity().as_datetime()),
        # MO value is unset, yield nothing
        (Unset(), []),
    ],
)
def test_get_changes_converts_symbolic_constants(input_mo_value, expected_mo_value):
    instance = _TestableCompareEndDate(
        None,  # mo_engagement_date_source, unused in this test
        None,  # ad_end_date_source, unused in this test
    )
    ad_user = ADUserEndDate(MO_UUID, ENDDATE_FIELD, VALID_AD_DATE)
    # Patch the `get_results` method to return our mock data, so we can test how it is
    # processed by `get_changes`.
    with patch.object(
        instance,
        "get_results",
        return_value=((ad_user, mo_value) for mo_value in [input_mo_value]),
    ):
        changes = list(instance.get_changes())
        if expected_mo_value == []:
            assert len(changes) == 0
        else:
            assert len(changes) == 1
            assert changes[0] == (ad_user, expected_mo_value)


@patch("integrations.ad_integration.ad_common.AD._create_session")
@pytest.mark.parametrize(
    "mock_ad_enddate_source,expected_result",
    [
        # If no matching AD user, don't return a MO user UUID and MO end date
        (_MockADEndDateSourceNoMatchingADUser(), {}),
        # If matching AD user exists *and* its AD end date is already up to date, don't
        # return a MO user UUID and MO end date.
        (_MockADEndDateSourceMatchingADUserAndEndDate(), {}),
        # If matching AD user exists *but* its AD end date is *blank*, return the MO
        # user UUID and MO end date.
        (_MockADEndDateSourceMatchingADUser(), {MO_UUID: VALID_AD_DATE}),
        # If matching AD user exists *but* its AD end date is *not up to date*, return
        # the MO user UUID and MO end date.
        (_MockADEndDateSourceMatchingADUserWrongEndDate(), {MO_UUID: VALID_AD_DATE}),
    ],
)
def test_get_changes_single_end_date(
    mock_create_session,  # patch(...)
    mock_mo_engagement_date_source: MOEngagementDateSource,  # pytest fixture
    mock_ad_enddate_source: ADEndDateSource,  # pytest parametrize, arg 0
    expected_result,  # pytest parametrize, arg 1
):
    instance = _TestableCompareEndDate(
        mock_mo_engagement_date_source,
        mock_ad_enddate_source,
    )
    actual_result = {
        ad_user_end_date.mo_uuid: new_end_date.strftime("%Y-%m-%d")
        for ad_user_end_date, new_end_date in instance.get_changes()
    }
    assert actual_result == expected_result


@patch("integrations.ad_integration.ad_common.AD._create_session")
def test_get_changes_handles_keyerror(
    mock_create_session,
    mock_mo_engagement_date_source_raising_keyerror,
):
    instance = _TestableCompareEndDate(
        mock_mo_engagement_date_source_raising_keyerror,
        _MockADEndDateSourceMatchingADUser(),
    )
    assert list(instance.get_changes()) == []


@pytest.mark.parametrize(
    "reader,enddate_field_future,expected_result,expected_normalized_value",
    [
        (MockADParameterReader(), None, [], "<unused>"),
        (
            MockADParameterReaderWithMOUUID(),
            None,
            [ADUserEndDate(MO_UUID, ENDDATE_FIELD, Invalid())],
            Invalid(),
        ),
        (
            MockADParameterReaderWithMOUUIDAndInvalidDate(),
            None,
            [ADUserEndDate(MO_UUID, ENDDATE_FIELD, INVALID_AD_DATE)],
            Invalid(),
        ),
        (
            MockADParameterReaderWithMOUUIDAndValidSplitDates(),
            ENDDATE_FIELD_FUTURE,
            [
                ADUserEndDate(MO_UUID, ENDDATE_FIELD, VALID_AD_DATE),
                ADUserEndDate(MO_UUID, ENDDATE_FIELD_FUTURE, VALID_AD_DATE_FUTURE),
            ],
            Invalid(),
        ),
    ],
)
def test_ad_end_date_source(
    reader: ADParameterReader,
    enddate_field_future: str | None,
    expected_result: list[ADUserEndDate],
    expected_normalized_value: datetime.datetime | Invalid,
):
    def assert_result_matches(actual_result, expected_result):
        assert actual_result == expected_result
        if actual_result:
            assert [
                actual.normalized_value == expected_normalized_value
                for actual in actual_result
            ]

    with patch(
        "integrations.ad_integration.ad_fix_enddate.ADParameterReader",
        return_value=reader,
    ):
        instance = ADEndDateSource(
            AD_UUID_FIELD,
            ENDDATE_FIELD,
            enddate_field_future,
            settings=TEST_SETTINGS,
        )

        assert_result_matches(list(instance.of_all_users()), expected_result)
        assert_result_matches(list(instance.of_one_user("username")), expected_result)


@patch("integrations.ad_integration.ad_common.AD._create_session")
@given(uuid=st.uuids(), enddate=st.dates())
def test_get_update_cmd(mock_session, uuid, enddate):
    u = _TestableUpdateEndDate()
    cmd = u.get_update_cmd(AD_UUID_FIELD, uuid, ENDDATE_FIELD, enddate)
    assert (
        cmd
        == f"""
        Get-ADUser  -SearchBase "{TEST_SEARCH_BASE}"  -Credential $usercredential -Filter \'{AD_UUID_FIELD} -eq "{uuid}"\' |
        Set-ADUser  -Credential $usercredential -Replace @{{{ENDDATE_FIELD}="{enddate}"}} |
        ConvertTo-Json
        """
    )


@patch("integrations.ad_integration.ad_common.AD._create_session")
@given(st.lists(st.tuples(st.builds(ADUserEndDate), st.datetimes()), max_size=10))
def test_run_all(mock_session, changes: list):
    u = _TestableUpdateEndDate()
    retval = u.run_all(changes, AD_UUID_FIELD)
    assert len(retval) == len(changes)
    for ps_script, ps_result in retval:
        assert "Set-ADUser" in ps_script
        assert ps_result == {}


@patch("integrations.ad_integration.ad_common.AD._create_session")
def test_run_all_logs_results_and_errors(mock_session, caplog):
    mock_changes = [
        (ADUserEndDate(MO_UUID, ENDDATE_FIELD, "2022-01-01"), dt("2022-06-01"))
    ]
    u = _TestableUpdateEndDateReturningError()
    with caplog.at_level(logging.DEBUG):
        u.run_all(mock_changes, AD_UUID_FIELD)
    assert len(caplog.records) == 5
    assert "Updating AD user" in caplog.records[0].message
    assert "Set-ADUser" in caplog.records[1].message  # command itself is logged
    assert "AD error response" in caplog.records[2].message  # error from AD is logged
    assert caplog.records[3].message == "0 users end dates corrected"
    assert caplog.records[4].message == "All end dates are fixed"


@patch("integrations.ad_integration.ad_common.AD._create_session")
@given(
    st.lists(
        st.tuples(st.builds(ADUserEndDate), st.datetimes()), min_size=1, max_size=1
    )
)
def test_run_all_supports_dry_run(mock_session, changes: list):
    u = _TestableUpdateEndDate()
    retval = u.run_all(changes, AD_UUID_FIELD, dry=True)
    assert len(retval) == len(changes)
    ps_script, ps_result = retval[0]
    assert "Set-ADUser" in ps_script
    assert ps_result == "<dry run>"
    assert len(u._ps_scripts_run) == 0


def mock_option_default(option, context, **kwargs):
    if option.name == "auth_server":
        return "http://keycloak"
    return None


@patch("click.core.Option.get_default", new=mock_option_default)
@patch("integrations.ad_integration.ad_common.read_settings", return_value={})
@patch("integrations.ad_integration.ad_common.AD._create_session")
@patch(
    "integrations.ad_integration.ad_fix_enddate.ADEndDateSource",
    cls=_MockADEndDateSourceNoMatchingADUser,
)
@pytest.mark.parametrize(
    "args",
    [
        [],
        ["--ad-user", "foobar"],
        ["--dry-run"],
    ],
)
def test_cli(
    mock_read_settings, mock_create_session, mock_ad_end_date_source, args, caplog
):
    runner = CliRunner()
    with caplog.at_level(logging.INFO):
        result = runner.invoke(cli, args)
    if result.exception is not None:
        raise result.exception
    assert caplog.records[0].message.startswith("Command line args: ")
    assert caplog.records[-1].message == "All end dates are fixed"



import asyncio
import enum
from typing import Iterable
from contextlib import contextmanager
from functools import partial

from fastapi import FastAPI
from prometheus_fastapi_instrumentator import Instrumentator
from prometheus_client import Enum
from prometheus_client import Gauge
from integrations.rundb.db_overview import DBOverview

from .config import get_changed_at_settings
from .sd_changed_at import changed_at


class State(enum.Enum):
    RUNNING = "running"
    OK = "ok"
    FAILURE = "failure"
    UNKNOWN = "unknown"


def get_state() -> State:
    try:
        settings = get_changed_at_settings()
        run_db = settings.sd_import_run_db
        db_overview = DBOverview(run_db)

        status_line = db_overview._read_last_line("status")

        if "Running since" in status_line:
            return State.RUNNING
        if "Update finished" in status_line:
            return State.OK

        return State.UNKNOWN
    except:
        return State.FAILURE


state = Enum(
    "sd_changed_at_state",
    "Current state of the SD Changed At integration",
    states=[s.value for s in State],
)
state.state(get_state().value)
start_time = Gauge(
    "sd_changed_at_start_time", "Start time of the latest SD Changed At run"
)
end_time = Gauge("sd_changed_at_end_time", "End time of the latest SD Changed At run")


app = FastAPI()
Instrumentator().instrument(app).expose(app)


@contextmanager
def update_state_metric() -> Iterable[None]:
    """Update SDChangedAt state metrics."""
    start_time.set_to_current_time()

    # TODO: write a test of this contextmanager
    # TODO: refactor the contextmanager to use the get_state function above
    settings = get_changed_at_settings()
    run_db = settings.sd_import_run_db
    db_overview = DBOverview(run_db)
    try:
        status = db_overview._read_last_line("status")
        state.state("running")
        yield
        if "Running since" in status:
            state.state("failure")
        elif "Update finished" in status:
            state.state("ok")
        else:
            state.state("unknown")
    except:
        state.state("failure")
        raise
    finally:
        end_time.set_to_current_time()


@app.get("/")
async def index() -> dict[str, str]:
    return {"name": "sdlon"}


@app.post("/trigger")
@app.get("/trigger", deprecated=True)
async def trigger(force: bool = False) -> dict[str, str]:
    loop = asyncio.get_running_loop()
    loop.call_soon(partial(update_state_metric()(changed_at), init=False, force=force))
    return {"triggered": "OK"}



import logging
import pickle
import typing
from typing import Tuple

import click
import sentry_sdk
from dateutil import tz

from .gql_lora_cache_async import GQLLoraCache
from .gql_lora_cache_async import GqlLoraCacheSettings
from .old_lora_cache import OldLoraCache

logger = logging.getLogger(__name__)

DEFAULT_TIMEZONE = tz.gettz("Europe/Copenhagen")

PICKLE_PROTOCOL = pickle.DEFAULT_PROTOCOL


def get_cache(resolve_dar=True, full_history=False, skip_past=False, settings=None):
    if GqlLoraCacheSettings().sentry_dsn:
        sentry_sdk.init(dsn=GqlLoraCacheSettings().sentry_dsn)

    if GqlLoraCacheSettings().use_new_cache:
        return GQLLoraCache(
            resolve_dar=resolve_dar,
            full_history=full_history,
            skip_past=skip_past,
            settings=settings,
        )
    else:
        return OldLoraCache(
            resolve_dar=resolve_dar,
            full_history=full_history,
            skip_past=skip_past,
            settings=settings,
        )


def fetch_loracache() -> Tuple[
    typing.Union[OldLoraCache, GQLLoraCache], typing.Union[OldLoraCache, GQLLoraCache]
]:
    # Here we should activate read-only mode, actual state and
    # full history dumps needs to be in sync.

    # Full history does not calculate derived data, we must
    # fetch both kinds.
    lc = get_cache(resolve_dar=True, full_history=False)
    lc.populate_cache(skip_associations=True)
    lc.calculate_derived_unit_data()
    lc.calculate_primary_engagements()

    # Todo, in principle it should be possible to run with skip_past True
    # This is now fixed in a different branch, remember to update when
    # merged.
    lc_historic = get_cache(resolve_dar=True, full_history=True, skip_past=False)
    lc_historic.populate_cache(skip_associations=True)
    # Here we should de-activate read-only mode
    return lc, lc_historic


@click.command()
@click.option("--historic/--no-historic", default=True, help="Do full historic export")
@click.option(
    "--skip-past", is_flag=True, default=False, help="Skip past in historic export"
)
@click.option(
    "--resolve-dar/--no-resolve-dar",
    default=False,
    envvar="RESOLVE_DAR",
    help="Resolve DAR addresses",
)
@click.option("--read-from-cache", is_flag=True)
def cli(historic, skip_past, resolve_dar, read_from_cache):
    GqlLoraCacheSettings().start_logging_based_on_settings()
    lc = get_cache(
        full_history=historic,
        skip_past=skip_past,
        resolve_dar=resolve_dar,
    )
    lc.populate_cache(dry_run=read_from_cache)

    logger.info("Now calcualate derived data")
    lc.calculate_derived_unit_data()
    lc.calculate_primary_engagements()


if __name__ == "__main__":

    cli()
